# coding=utf-8
"""Base class for all 1D geometries in 3D space (Ray3D and LineSegment3D)."""
from __future__ import division

from .pointvector import Point3D, Vector3D
from ..intersection3d import closest_point3d_on_line3d, \
    closest_point3d_on_line3d_infinite, intersect_line3d_plane


class Base1DIn3D(object):
    """Base class for all 1D geometries in 3D space (Ray3D and LineSegment3D).

    Args:
        p: A Point2D representing the base.
        v: A Vector2D representing the direction.

    Properties:
        * p
        * v
    """
    __slots__ = ('_p', '_v')

    def __init__(self, p, v):
        """Initilize Base1DIn3D."""
        assert isinstance(p, Point3D), "Expected Point3D. Got {}.".format(type(p))
        assert isinstance(v, Vector3D), "Expected Vector3D. Got {}.".format(type(v))
        self._p = p
        self._v = v

    @classmethod
    def from_dict(cls, data):
        """Create a LineSegment3D/Ray3D from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "p": (10, 0, 0),
            "v": (10, 10, 0)
            }
        """
        return cls(Point3D.from_array(data['p']),
                   Vector3D.from_array(data['v']))

    @property
    def p(self):
        """Base point."""
        return self._p

    @property
    def v(self):
        """Direction vector."""
        return self._v

    def is_parallel(self, line_ray, angle_tolerance):
        """Test whether this object is parallel to another LineSegment3D or Ray3D.

        Args:
            line_ray: Another LineSegment3D or Ray3D for which parallelization
                with this objects will be tested.
            angle_tolerance: The max angle in radians that the direction between
                this object and another can vary for them to be considered
                parallel.
        """
        if self.v.angle(line_ray.v) <= angle_tolerance:
            return True
        elif self.v.angle(line_ray.v.reverse()) <= angle_tolerance:
            return True
        return False

    def is_colinear(self, line_ray, tolerance, angle_tolerance):
        """Test whether this object is colinear to another LineSegment3D or Ray3D.

        Args:
            line_ray: Another LineSegment3D or Ray3D for which colinearity
                with this object will be tested.
            tolerance: The maximum distance between the line_ray and the infinite
                extension of this object for them to be cinsidered colinear.
            angle_tolerance: The max angle in radians that the direction between
                this object and another can vary for them to be considered
                parallel.
        """
        if not self.is_parallel(line_ray, angle_tolerance):
            return False
        _close_pt = closest_point3d_on_line3d_infinite(self.p, line_ray)
        if self.p.distance_to_point(_close_pt) >= tolerance:
            return False
        return True

    def closest_point(self, point):
        """Get the closest Point3D on this object to another Point3D.

        Args:
            point: A Point3D object to which the closest point on this object
                will be computed.

        Returns:
            Point3D for the closest point on this line/ray to the input point.
        """
        return closest_point3d_on_line3d(point, self)

    def distance_to_point(self, point):
        """Get the minimum distance between this object and the input point.

        Args:
            point: A Point3D object to which the minimum distance will be computed.

        Returns:
            The distance to the input point.
        """
        close_pt = self.closest_point(point)
        return point.distance_to_point(close_pt)

    def intersect_plane(self, plane):
        """Get the intersection between this object and a Plane.

        Args:
            plane: A Plane that will be intersected with this object.

        Returns:
            A Point3D object if the intersection was successful.
            None if no intersection exists.
        """
        return intersect_line3d_plane(self, plane)

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def to_dict(self):
        """Get LineSegment3D/Ray3D as a dictionary."""
        return {'p': self.p.to_array(),
                'v': self.v.to_array()}

    def __copy__(self):
        return self.__class__(self.p, self.v)
    
    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (hash(self.p), hash(self.v))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Base1DIn3D) and self.__key() == other.__key()
    
    def __ne__(self, other):
        return not self.__eq__(other)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Base1Din3D representation."""
        return 'Base 1D Object (3D Space)'



# coding=utf-8
"""Object with Multiple Planar Faces in 3D Space"""
from __future__ import division

from .pointvector import Vector3D, Point3D
from .ray import Ray3D
from .line import LineSegment3D
from .plane import Plane
from .face import Face3D
from ._2d import Base2DIn3D

try:
    from itertools import izip as zip  # python 2
except ImportError:
    xrange = range  # python 3


class Polyface3D(Base2DIn3D):
    """Object with Multiple Planar Faces in 3D Space. Includes solid objects and polyhedra.

    Args:
        vertices: A list of Point3D objects representing the vertices of
            this PolyFace.
        face_indices: A list of lists with one list for each face of the polyface.
            Each face list must contain at least one tuple of integers corresponding
            to indices within the vertices list. Additional tuples of integers may
            follow this one such that the first tuple denotes the boundary of the
            face while each subsequent tuple denotes a hole in the face.
        edge_information: Optional edge information, which will speed up the
            creation of the Polyface object if it is available but should be left
            as None if it is unknown. If None, edge_information will be computed
            from the vertices and face_indices inuputs. Edge information
            should be formatted as a dictionary with two keys as follows:

            *   'edge_indices':
                An array objects that each contain two integers.
                These integers correspond to indices within the vertices list and
                each tuple represents a line sengment for an edge of the polyface.
            *   'edge_types':
                An array of integers for each edge that parallels the edge_indices
                list. An integer of 0 denotes a naked edge, an integer of 1
                denotes an internal edge. Anything higher is a non-manifold edge.

    Properties:
        * vertices
        * faces
        * edges
        * naked_edges
        * internal_edges
        * non_manifold_edges
        * face_indices
        * edge_indices
        * edge_types
        * min
        * max
        * center
        * area
        * volume
        * is_solid
    """
    __slots__ = ('_faces', '_edges',
                 '_naked_edges', '_internal_edges', '_non_manifold_edges',
                 '_face_indices', '_edge_indices', '_edge_types',
                 '_area', '_volume', '_is_solid')

    def __init__(self, vertices, face_indices, edge_information=None):
        """Initilize Polyface3D."""
        # assign input properties
        Base2DIn3D.__init__(self, vertices)
        self._face_indices = tuple(tuple(tuple(loop) for loop in face)
                                   for face in face_indices)

        if edge_information is not None:  # unpack the input edge information
            edge_i = edge_information['edge_indices']
            edge_t = edge_information['edge_types']
        else:  # autocalculate the edge information from the vertices and faces
            edge_i = []
            edge_t = []
            for face in face_indices:
                for fi in face:
                    for i, vi in enumerate(fi):
                        try:  # this can get slow for large number of vertices
                            ind = edge_i.index((vi, fi[i - 1]))
                            edge_t[ind] += 1
                        except ValueError:  # make sure reversed edge isn't there
                            try:
                                ind = edge_i.index((fi[i - 1], vi))
                                edge_t[ind] += 1
                            except ValueError:  # add a new edge
                                edge_i.append((fi[i - 1], vi))
                                edge_t.append(0)
        self._edge_indices = edge_i if isinstance(edge_i, tuple) else tuple(edge_i)
        self._edge_types = edge_t if isinstance(edge_t, tuple) else tuple(edge_t)

        # determine solidity of the polyface by checking for internal edges
        self._is_solid = True
        for edge in self._edge_types:
            if edge != 1:
                self._is_solid = False
                break

        # assign default properties
        self._faces = None
        self._edges = None
        self._naked_edges = None
        self._internal_edges = None
        self._non_manifold_edges = None
        self._area = None
        self._volume = None

    @classmethod
    def from_dict(cls, data):
        """Create a Face3D from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "Polyface3D",
            "vertices": [(0, 0, 0), (10, 0, 0), (10, 10, 0), (0, 10, 0)],
            "face_indices": [[(0, 1, 2)], [(3, 0, 1)]],
            "edge_information": {"edge_indices":[(0, 1), (1, 2), (2, 0), (2, 3), (3, 0)],
                                 "edge_types":[0, 0, 1, 0, 0]}
            }
        """
        if 'edge_information' in data and data['edge_information'] is not None:
            edge_information = data['edge_information']
        else:
            edge_information = None

        return cls(tuple(Point3D.from_array(pt) for pt in data['vertices']),
                   data['face_indices'], edge_information)

    @classmethod
    def from_faces(cls, faces, tolerance=0):
        """Initilize Polyface3D from a list of Face3D objects.

        Note that the Polyface3D.faces property of the resulting polyface will
        have an order of faces that matches the order input to this classmethod.

        Args:
            faces: A list of Face3D objects representing the boundary of this Polyface.
            tolerance: The maximum difference between x, y, and z values at which
                the vertex of two adjacent faces is considered the same. Default: 0.
        """
        # extract unique vertices from the faces
        vertices = []  # collection of vertices as point objects
        face_indices = []  # collection of face indices
        for f in faces:
            ind = []
            loops = (f.boundary,) if not f.has_holes else (f.boundary,) + f.holes
            for j, loop in enumerate(loops):
                ind.append([])
                for v in loop:
                    found = False
                    for i, vert in enumerate(vertices):
                        if v.is_equivalent(vert, tolerance):
                            found = True
                            ind[j].append(i)
                            break
                    if not found:  # add new point
                        vertices.append(v)
                        ind[j].append(len(vertices) - 1)
            face_indices.append(tuple(ind))

        # get the polyface object and assign correct faces to it
        return cls(vertices, face_indices)

    @classmethod
    def from_box(cls, width, depth, height, base_plane=None):
        """Initilize Polyface3D from parameters describing a box.

        Initializing a polyface this way has the added benefit of having its
        faces property quickly calculated.

        Args:
            width: A number for the width of the box (in the X direction).
            depth: A number for the depth of the box (in the Y direction).
            height: A number for the height of the box (in the Z direction).
            base_plane: A Plane object from which to generate the box.
                If None, default is the WorldXY plane.
        """
        assert isinstance(width, (float, int)), 'Box width must be a number.'
        assert isinstance(depth, (float, int)), 'Box depth must be a number.'
        assert isinstance(height, (float, int)), 'Box height must be a number.'
        if base_plane is not None:
            assert isinstance(base_plane, Plane), \
                'base_plane must be Plane. Got {}.'.format(type(base_plane))
        else:
            base_plane = Plane(Vector3D(0, 0, 1), Point3D())
        _o = base_plane.o
        _w_vec = base_plane.x * width
        _d_vec = base_plane.y * depth
        _h_vec = base_plane.n * height
        _verts = (_o, _o + _d_vec, _o + _d_vec + _w_vec, _o + _w_vec,
                  _o + _h_vec, _o + _d_vec + _h_vec,
                  _o + _d_vec + _w_vec + _h_vec, _o + _w_vec + _h_vec)
        _face_indices = ([(0, 1, 2, 3)], [(2, 1, 5, 6)], [(6, 7, 3, 2)],
                         [(0, 3, 7, 4)], [(0, 4, 5, 1)], [(7, 6, 5, 4)])
        _edge_indices = ((3, 0), (0, 1), (1, 2), (2, 3), (0, 4), (4, 5),
                         (5, 1), (3, 7), (7, 4), (6, 2), (5, 6), (6, 7))
        polyface = cls(_verts, _face_indices, {'edge_indices': _edge_indices,
                                               'edge_types': [1] * 12})
        verts = tuple(tuple(_verts[i] for i in face[0]) for face in _face_indices)
        bottom = Face3D(verts[0], base_plane.flip(), enforce_right_hand=False)
        middle = tuple(Face3D(v, enforce_right_hand=False) for v in verts[1:5])
        top = Face3D(verts[5], base_plane.move(_h_vec), enforce_right_hand=False)
        polyface._faces = (bottom,) + middle + (top,)
        polyface._volume = width * depth * height
        return polyface

    @classmethod
    def from_offset_face(cls, face, offset):
        """Initilize a solid Polyface3D from a Face3D offset along its normal.

        The resulting polyface will always be offset in the direction of
        the face normal.

        When a polyface is initialized this way, the first face of the
        Polysurface3D.faces will always be the input face used to create the
        object, the last face will be the offset version of the face, and all
        other faces will form the extrusion connecting the two.

        Args:
            face: A Face3D to serve as a base for the polyface.
            offset: A number for the distance to offset the face to make a solid.
        """
        assert isinstance(face, Face3D), \
            'face must be a Face3D. Got {}.'.format(type(face))
        assert isinstance(offset, (float, int)), \
            'height must be a number. Got {}.'.format(type(offset))
        # compute vertices, face indices, and edges of the extrusion
        extru_vec = face.normal * offset
        verts, face_ind_extru, edge_indices = \
            Polyface3D._verts_faces_edges_from_boundary(face.boundary, extru_vec)
        if face.has_holes:
            _st_i = len(verts)
            for i, hole in enumerate(face.hole_polygon2d):
                hole_verts = face._holes[i] if hole.is_clockwise else \
                    tuple(reversed(face._holes[i]))
                verts_2, face_ind_extru_2, edge_indices_2 = \
                    Polyface3D._verts_faces_edges_from_boundary(
                        hole_verts, extru_vec, _st_i)
                verts.extend(verts_2)
                face_ind_extru.extend(face_ind_extru_2)
                edge_indices.extend(edge_indices_2)
                _st_i += len(hole_verts)
        face_ind_extru = [[fc] for fc in face_ind_extru]
        # compute the final faces (accounting for top and bottom)
        if not face.has_holes:
            len_faces = len(face.boundary)
            face_ind_bottom = [tuple(reversed(xrange(len_faces)))]
            face_ind_top = [tuple(
                reversed(xrange(len_faces * 2 - 1, len_faces - 1, -1)))]
        else:
            face_verts_bottom = [list(reversed(face.boundary))] + list(face.holes)
            face_verts_top = [[pt.move(extru_vec) for pt in reversed(loop)]
                              for loop in face_verts_bottom]
            face_ind_bottom = [tuple(verts.index(pt) for pt in loop)
                               for loop in face_verts_bottom]
            face_ind_top = [tuple(verts.index(pt) for pt in loop)
                            for loop in face_verts_top]
        faces_ind = [face_ind_bottom] + face_ind_extru + [face_ind_top]
        # create the polysurface and assign known properties.
        polyface = cls(verts, faces_ind, {'edge_indices': edge_indices,
                                          'edge_types': [1] * len(edge_indices)})
        polyface._volume = face.area * offset
        face_verts = tuple(
            tuple(tuple(verts[i] for i in loop) for loop in f) for f in faces_ind)
        if not face.has_holes:
            polyface._faces = tuple(Face3D(v[0], enforce_right_hand=False)
                                    for v in face_verts)
        else:
            mid_faces = [Face3D(v[0], enforce_right_hand=False)
                         for v in face_verts[1:-1]]
            bottom_face = face.flip()
            top_face = face.move(extru_vec)
            polyface._faces = tuple([bottom_face] + mid_faces + [top_face])
        return polyface

    @property
    def vertices(self):
        """Tuple of all vertices in this polyface.

        Note that, in the case of a polyface with holes, some vertices will be repeated
        since this property effectively traces out a single boundary around the
        whole shape, winding inward to cut out the holes.
        """
        return self._vertices

    @property
    def faces(self):
        """Tuple of all Face3D objects making up this polyface."""
        if self._faces is None:
            faces = []
            for face in self._face_indices:
                boundary = tuple(self.vertices[i] for i in face[0])
                if len(face) == 1:
                    faces.append(Face3D(boundary))
                else:
                    holes = tuple(tuple(self.vertices[i] for i in f) for f in face[1:])
                    faces.append(Face3D(boundary=boundary, holes=holes))
            if self._is_solid:
                self._faces = Polyface3D.get_outward_faces(faces)
            else:
                self._faces = tuple(faces)
        return self._faces

    @property
    def edges(self):
        """"Tuple of all edges in this polyface as LineSegment3D objects."""
        if self._edges is None:
            self._edges = tuple(LineSegment3D.from_end_points(
                self.vertices[seg[0]], self.vertices[seg[1]])
                                for seg in self._edge_indices)
        return self._edges

    @property
    def naked_edges(self):
        """"Tuple of all naked edges in this polyface as LineSegment3D objects.

        Naked edges belong to only one face in the polyface (they are not
        shared between faces).
        """
        if self._naked_edges is None:
            self._naked_edges = self._get_edge_type(0)
        return self._naked_edges

    @property
    def internal_edges(self):
        """"Tuple of all internal edges in this polyface as LineSegment3D objects.

        Internal edges are shared between two faces in the polyface.
        """
        if self._internal_edges is None:
            self._internal_edges = self._get_edge_type(1)
        return self._internal_edges

    @property
    def non_manifold_edges(self):
        """"Tuple of all non-manifold edges in this polyface as LineSegment3D objects.

        Non-manifold edges are shared between three or more faces and are therefore
        not allowed in solid polyfaces.
        """
        if self._non_manifold_edges is None:
            if self._edges is None:
                self.edges
            nm_edges = []
            for i, type in enumerate(self._edge_types):
                if type > 2:
                    nm_edges.append(self._edges[i])
            self._non_manifold_edges = tuple(nm_edges)
        return self._non_manifold_edges

    @property
    def face_indices(self):
        """Tuple of face tuples with integers corresponding to indices of vertices."""
        return self._face_indices

    @property
    def edge_indices(self):
        """Tuple of edge tuples with integers corresponding to indices of vertices."""
        return self._edge_indices

    @property
    def edge_types(self):
        """Tuple of integers for each edge that denotes the type of edge.

        0 denotes a naked edge, 1 denotes an internal edge, and anything higher is a
        non-manifold edge.
        """
        return self._edge_types

    @property
    def edge_information(self):
        """Dictionary with keys: 'edge_indices', 'edge_types' and corresponding properties.
        """
        return {'edge_indices': self._edge_indices, 'edge_types': self._edge_types}

    @property
    def area(self):
        """The total surface area of the polyface."""
        if self._area is None:
            self._area = sum([face.area for face in self.faces])
        return self._area

    @property
    def volume(self):
        """The volume enclosed by the polyface.

        Note that, if this polyface is not solid (with all face normals pointing
        outward), the value of this property will not be valid.
        """
        if self._volume is None:
            # formula taken from https://en.wikipedia.org/wiki/Polyhedron#Volume
            _v = 0
            for i, face in enumerate(self.faces):
                _v += face[0].dot(face.normal) * face.area
            self._volume = _v / 3
        return self._volume

    @property
    def is_solid(self):
        """A boolean to note whether the polyface is solid (True) or is open (False).

        Note that all solid polyface objects will have faces pointing outwards.
        """
        return self._is_solid

    def merge_overlapping_edges(self, tolerance, angle_tolerance):
        """Get this object with overlapping naked edges merged into single internal edges.

        This can be used to determine if a polyface is truly solid.
        The default test of edge conditions that runs upon creation of a polyface does
        not check for cases where overlapping colinear edges could be considered
        a single internal edge such as the case below:

        .. code-block:: shell

                             |           1          |
                            A|______________________|C
                             |          B|          |
                             |           |          |
                             |     2     |     3    |

        If Face 1 only has edge AC and not two separate edges for AB and BC, the
        creation of the polyface will yield naked edges for AC, AB, and BC, meaning
        the shape would not be considered solid when it might actually be so. This
        merge_overlapping_edges method overcomes this by replacing the entire set
        of 3 naked edges above a single internal edge running from A to C.

        Args:
            tolerance: The minimum distance between a vertex and the boundary segments
                at which point the vertex is considered colinear.
            angle_tolerance: The max angle in radians that vertices are allowed to
                 differ from one another in order to consider them colinear.
        """
        # get naked edges
        naked_edges = list(self.naked_edges)
        if len(naked_edges) == 0:
            return self

        # establish lists that will be iteratively edited
        remove_i = []
        add_edges = []
        naked_edge_i = []
        naked_edge_ind = []
        for i, x in enumerate(self.edge_types):
            if x == 0:
                naked_edge_i.append(i)
                naked_edge_ind.append(self._edge_indices[i])

        while len(naked_edge_i) > 1:
            # get all of the edges that are colinear with the first edge
            coll_edges = list(naked_edge_ind[0])
            coll_i = [naked_edge_i[0]]
            kept_i = []
            for edge, ind, nei, i in zip(
                    naked_edges[1:], naked_edge_ind[1:], naked_edge_i[1:],
                    xrange(1, len(naked_edges))):
                if edge.is_colinear(naked_edges[0], tolerance, angle_tolerance):
                    coll_edges.extend(ind)
                    coll_i.append(nei)
                else:
                    kept_i.append(i)

            # determine if  colinear edges create a full double line along the edge
            if len(coll_edges) == 1:
                overlapping = False
            else:
                final_vi = []
                coll_edges.sort()
                overlapping = True
                for i in range(0, len(coll_edges), 2):
                    final_vi.append(coll_edges[i])
                    if not coll_edges[i] == coll_edges[i + 1]:
                        overlapping = False
                        break

            # if fully overlapping edges have been found, remake them into one
            if overlapping:
                remove_i.extend(coll_i)  # remove overlapping edges from the list
                verts = [self.vertices[j] for j in final_vi]
                dir_vec = verts[0] - verts[1]
                if dir_vec.x != 0:
                    vert_coor = [v.x for v in verts]
                elif dir_vec.y != 0:
                    vert_coor = [v.y for v in verts]
                else:
                    vert_coor = [v.z for v in verts]
                vert_coor, final_vi = zip(*sorted(zip(vert_coor, final_vi)))
                add_edges.append((final_vi[0], final_vi[-1]))

            # delete the colinear vertices that have been accounted for
            naked_edges = [naked_edges[i] for i in kept_i]
            naked_edge_ind = [naked_edge_ind[i] for i in kept_i]
            naked_edge_i = [naked_edge_i[i] for i in kept_i]

        # create the new edge information and the new polyface
        new_edge_indices = list(self.edge_indices)
        new_edge_types = list(self.edge_types)
        add_i = []
        for i in range(len(new_edge_indices)):
            if i not in remove_i:
                add_i.append(i)
        new_edge_indices = [new_edge_indices[i] for i in add_i]
        new_edge_types = [new_edge_types[i] for i in add_i]
        for new_edge in add_edges:
            new_edge_indices.append(new_edge)
            new_edge_types.append(1)
        _new_polyface = Polyface3D(self._vertices, self._face_indices,
                                   {'edge_indices': new_edge_indices,
                                    'edge_types': new_edge_types})
        return _new_polyface

    def move(self, moving_vec):
        """Get a polyface that has been moved along a vector.

        Args:
            moving_vec: A Vector3D with the direction and distance to move the polyface.
        """
        _verts = tuple(pt.move(moving_vec) for pt in self.vertices)
        _new_pface = Polyface3D(_verts, self.face_indices, self.edge_information)
        if self._faces is not None:
            _new_pface._faces = tuple(face.move(moving_vec) for face in self._faces)
        _new_pface._volume = self._volume
        return _new_pface

    def rotate(self, axis, angle, origin):
        """Rotate a polyface by a certain angle around an axis and origin.

        Right hand rule applies:
        If axis has a positive orientation, rotation will be clockwise.
        If axis has a negative orientation, rotation will be counterclockwise.

        Args:
            axis: A Vector3D axis representing the axis of rotation.
            angle: An angle for rotation in radians.
            origin: A Point3D for the origin around which the object will be rotated.
        """
        _verts = tuple(pt.rotate(axis, angle, origin) for pt in self.vertices)
        _new_pface = Polyface3D(_verts, self.face_indices, self.edge_information)
        if self._faces is not None:
            _new_pface._faces = tuple(face.rotate(axis, angle, origin)
                                      for face in self._faces)
        _new_pface._volume = self._volume
        return _new_pface

    def rotate_xy(self, angle, origin):
        """Get a polyface rotated counterclockwise in the world XY plane by an angle.

        Args:
            angle: An angle in radians.
            origin: A Point3D for the origin around which the object will be rotated.
        """
        _verts = tuple(pt.rotate_xy(angle, origin) for pt in self.vertices)
        _new_pface = Polyface3D(_verts, self.face_indices, self.edge_information)
        if self._faces is not None:
            _new_pface._faces = tuple(face.rotate_xy(angle, origin)
                                      for face in self._faces)
        _new_pface._volume = self._volume
        return _new_pface

    def reflect(self, normal, origin):
        """Get a polyface reflected across a plane with the input normal vector and origin.

        Args:
            normal: A Vector3D representing the normal vector for the plane across
                which the polyface will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point3D representing the origin from which to reflect.
        """
        _verts = tuple(pt.reflect(normal, origin) for pt in self.vertices)
        _new_pface = Polyface3D(_verts, self.face_indices, self.edge_information)
        if self._faces is not None:
            _new_pface._faces = tuple(face.reflect(normal, origin)
                                      for face in self._faces)
        _new_pface._volume = self._volume
        return _new_pface

    def scale(self, factor, origin=None):
        """Scale a polyface by a factor from an origin point.

        Args:
            factor: A number representing how much the polyface should be scaled.
            origin: A Point3D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0, 0).
        """
        if origin is None:
            _verts = tuple(Point3D(pt.x * factor, pt.y * factor, pt.z * factor)
                           for pt in self._vertices)
        else:
            _verts = tuple(pt.scale(factor, origin) for pt in self.vertices)
        _new_pface = Polyface3D(_verts, self.face_indices, self.edge_information)
        if self._faces is not None:
            _new_pface._faces = tuple(face.scale(factor, origin)
                                      for face in self._faces)
        _new_pface._volume = self._volume * factor ** 3 \
            if self._volume is not None else None
        return _new_pface

    def is_point_inside(self, point, test_vector=Vector3D(1, 0, 0)):
        """Test whether a Point3D lies inside or outside the polyface.

        Note that, if this polyface is not solid, the result will always be False.

        Args:
            point: A Point3D for which the inside/outside relationship will be tested.
            test_vector: Optional vector to set the direction in which intersections
                with the polyface faces will be evaluated to determine if the
                point is inside. Default is the X-unit vector.

        Returns:
            A boolean denoting whether the point lies inside (True) or outside (False).
        """
        if not self.is_solid:
            return False
        test_ray = Ray3D(point, test_vector)
        n_int = 0
        for _f in self.faces:
            if _f.intersect_line_ray(test_ray):
                n_int += 1
        if n_int % 2 == 0:
            return False
        return True

    def does_intersect_line_ray_exist(self, line_ray):
        """Boolean denoting whether an intersection exists between the input Line3D or Ray3D.

        Args:
            line_ray: A Line3D or Ray3D object for which intersection will be evaluated.

        Returns:
            True if an intersection exists. False if it does not exist.
        """
        for face in self.faces:
            _int = face.intersect_line_ray(line_ray)
            if _int is not None:
                return True
        return False

    def intersect_line_ray(self, line_ray):
        """Get the intersections between this polyface and the input Line3D or Ray3D.

        Args:
            line_ray: A Line3D or Ray3D object for which intersection will be computed.

        Returns:
            A list of Point3D for the intersection. Will be an empty list if no
            intersection exists.
        """
        _inters = []
        for face in self.faces:
            _int = face.intersect_line_ray(line_ray)
            if _int is not None:
                _inters.append(_int)
        return _inters

    def intersect_plane(self, plane):
        """Get the intersection between this polyface and the input plane.

        Args:
            plane: A Plane object for which intersection will be computed.

        Returns:
            List of LineSegment3D objects for the intersection.
            Will be an empty list if no intersection exists.
        """
        _inters = []
        for face in self.faces:
            _int = face.intersect_plane(plane)
            if _int is not None:
                _inters.extend(_int)
        return _inters

    @staticmethod
    def overlapping_bounding_boxes(polyface1, polyface2, tolerance):
        """Check if the bounding boxes of two polyfaces overlap within a tolerance.

        This is particularly useful as a check before performing computationally
        intenseprocesses between two polyfaces like intersection or checking for
        adjacency. Checking the overlap of the bounding boxes is extremely quick
        given this method's use of the Separating Axis Theorem.

        Args:
            polyface1: The first polyface to check.
            polyface2: The second polyface to check.
            tolerance: Distance within which two points are considered to be co-located.
        """
        # Bounding box check using the Separating Axis Theorem
        polyf1_width = polyface1.max.x - polyface1.min.x
        polyf2_width = polyface2.max.x - polyface2.min.x
        dist_btwn_x = abs(polyface1.center.x - polyface2.center.x)
        x_gap_btwn_box = dist_btwn_x - (0.5 * polyf1_width) - (0.5 * polyf2_width)

        polyf1_depth = polyface1.max.y - polyface1.min.y
        polyf2_depth = polyface2.max.y - polyface2.min.y
        dist_btwn_y = abs(polyface1.center.y - polyface2.center.y)
        y_gap_btwn_box = dist_btwn_y - (0.5 * polyf1_depth) - (0.5 * polyf2_depth)

        polyf1_height = polyface1.max.z - polyface1.min.z
        polyf2_height = polyface2.max.z - polyface2.min.z
        dist_btwn_z = abs(polyface1.center.z - polyface2.center.z)
        z_gap_btwn_box = dist_btwn_z - (0.5 * polyf1_height) - (0.5 * polyf2_height)

        if x_gap_btwn_box > tolerance or y_gap_btwn_box > tolerance or \
                z_gap_btwn_box > tolerance:
            return False  # no overlap
        return True  # overlap exists

    @staticmethod
    def get_outward_faces(faces, tolerance=0):
        """Get faces that are all pointing outward from a list of faces together forming a solid.

        Note that, if the input faces do not form a closed solid, there may be some
        output faces that are not pointing outward.  However, if the gaps in the
        combined solid are within the input tolerance, this should not be an issue.

        Also, note that this method runs automatically for any solid polyface
        (meaning every solid polyface automatically has outward-facing faces). So there
        is no need to rerun this method for faces from a solid polyface.

        Args:
            faces: A list of Face3D objects that together form a solid.
            tolerance: Optional tolerance for the permissable size of gap between
                faces at which point the faces are considered to have a single edge.

        Returns:
            outward_faces -- A list of the input Face3D objects that all point outwards
            (provided the input faces form a solid).
        """
        outward_faces = []
        for i, face in enumerate(faces):
            # construct a ray with the face normal and a point on the face
            v1 = face.boundary[-1] - face.boundary[0]
            v2 = face.boundary[1] - face.boundary[0]
            move_vec = Vector3D(
                (v1.x + v2.x / 2), (v1.y + v2.y / 2), (v1.z + v2.z / 2)).normalize()
            move_vec = move_vec * (tolerance + 0.00001)
            point_on_face = face.boundary[0] + move_vec
            test_ray = Ray3D(point_on_face, face.normal)

            # if the ray intersects with an even number of other faces, it is correct
            n_int = 0
            for _f in faces[i + 1:]:
                if _f.intersect_line_ray(test_ray):
                    n_int += 1
            for _f in faces[:i]:
                if _f.intersect_line_ray(test_ray):
                    n_int += 1
            if n_int % 2 == 0:
                outward_faces.append(face)
            else:
                outward_faces.append(face.flip())
        return outward_faces

    def to_dict(self, include_edge_information=True):
        """Get Polyface3D as a dictionary.

        Args:
            include_edge_information: Set to True to include the edge_information
                in the dictionary, which will allow for fast initialization when
                it is de-serialized. Default True.
        """
        base = {'type': 'Polyface3D',
                'vertices': [v.to_array() for v in self.vertices],
                'face_indices': self.face_indices}
        if include_edge_information:
            base['edge_information'] = self.edge_information
        return base

    def _get_edge_type(self, edge_type):
        """Get all of the edges of a certain type in this polyface."""
        if self._edges is None:
            self.edges
        sel_edges = []
        for i, type in enumerate(self._edge_types):
            if type == edge_type:
                sel_edges.append(self._edges[i])
        return tuple(sel_edges)

    @staticmethod
    def _verts_faces_edges_from_boundary(cclock_verts, extru_vec, st_i=0):
        verts = list(cclock_verts) + [pt.move(extru_vec) for pt in cclock_verts]
        len_faces = len(cclock_verts)
        faces_ind = []
        for i in xrange(st_i, st_i + len_faces - 1):
            faces_ind.append((i, i + 1, i + len_faces + 1, i + len_faces))
        faces_ind.append((st_i + len_faces - 1, st_i,
                          st_i + len_faces, st_i + len_faces * 2 - 1))
        edge_i1 = [(st_i + i, st_i + i + 1) for i in xrange(len_faces - 1)]
        edge_i2 = [(st_i + i, st_i + i + len_faces) for i in xrange(len_faces)]
        edge_i3 = [(st_i + len_faces + i, st_i + len_faces + i + 1)
                   for i in xrange(len_faces - 1)]
        edge_indices = edge_i1 + [(st_i + len_faces - 1, st_i)] + edge_i2 + edge_i3 + \
            [(st_i + len_faces * 2 - 1, st_i + len_faces)]
        return verts, faces_ind, edge_indices

    def __copy__(self):
        _new_poly = Polyface3D(self.vertices, self.face_indices, self.edge_information)
        _new_poly._faces = self._faces
        return _new_poly
    
    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return tuple(hash(pt) for pt in self._vertices) + \
            tuple(hash(face) for face in self._face_indices)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Polyface3D) and self.__key() == other.__key()

    def __repr__(self):
        return 'Polyface3D ({} faces) ({} vertices)'.format(
            len(self.faces), len(self))



# coding=utf-8
"""2D Polyline"""
from __future__ import division

from ._2d import Base2DIn2D
from .pointvector import Point2D
from .line import LineSegment2D
from .polygon import Polygon2D
from ..intersection2d import intersect_line2d, intersect_line2d_infinite
from .._polyline import _group_vertices


class Polyline2D(Base2DIn2D):
    """2D polyline object.

    Args:
        vertices: A list of Point2D objects representing the vertices of the polyline.
        interpolated: Boolean to note whether the polyline should be interpolated
            between the input vertices when it is translated to other interfaces.
            Note that this property has no bearing on the geometric calculations
            performed by this library and is only present in order to assist with
            display/translation.

    Properties:
        * vertices
        * segments
        * min
        * max
        * center
        * length
        * is_self_intersecting
        * interpolated
    """
    __slots__ = ('_interpolated', '_segments', '_length', '_is_self_intersecting')

    def __init__(self, vertices, interpolated=False):
        """Initilize Polyline2D."""
        Base2DIn2D.__init__(self, vertices)
        self._interpolated = interpolated
        self._segments = None
        self._length = None
        self._is_self_intersecting = None

    @classmethod
    def from_dict(cls, data):
        """Create a Polyline2D from a dictionary.

        Args:
            data: A python dictionary in the following format.

        .. code-block:: python

            {
            "type": "Polyline2D",
            "vertices": [(0, 0), (10, 0), (0, 10)]
            }
        """
        interp = data['interpolated'] if 'interpolated' in data else False
        return cls(tuple(Point2D.from_array(pt) for pt in data['vertices']), interp)

    @classmethod
    def from_array(cls, point_array):
        """Create a Polyline2D from a nested array of vertex coordinates.

        Args:
            point_array: Nested array of point arrays.
        """
        return Polyline2D(Point2D(*point) for point in point_array)

    @classmethod
    def from_polygon(cls, polygon):
        """Create a closed Polyline2D from a Polygon2D.

        Args:
            polygon: A Polygon2D object to be converted to a Polyline2D.
        """
        return Polyline2D(polygon.vertices + (polygon.vertices[0],))

    @property
    def segments(self):
        """Tuple of all line segments in the polyline."""
        if self._segments is None:
            self._segments = \
                tuple(LineSegment2D.from_end_points(vert, self._vertices[i + 1])
                      for i, vert in enumerate(self._vertices[:-1]))
        return self._segments

    @property
    def length(self):
        """The length of the polyline."""
        if self._length is None:
            self._length = sum([seg.length for seg in self.segments])
        return self._length

    @property
    def is_self_intersecting(self):
        """Boolean noting whether the polyline has self-intersecting segments."""
        if self._is_self_intersecting is None:
            self._is_self_intersecting = False
            _segs = self.segments
            for i, _s in enumerate(_segs[1: len(_segs) - 1]):
                _skip = (i, i + 1, i + 2)
                _other_segs = [x for j, x in enumerate(_segs) if j not in _skip]
                for _oth_s in _other_segs:
                    if _s.intersect_line_ray(_oth_s) is not None:  # intersection!
                        self._is_self_intersecting = True
                        break
                if self._is_self_intersecting is True:
                    break
        return self._is_self_intersecting

    @property
    def interpolated(self):
        """Boolean noting whether the polyline should be interpolated upon translation.

        Note that this property has no bearing on the geometric calculations
        performed by this library and is only present in order to assist with
        display/translation.
        """
        return self._interpolated

    def is_closed(self, tolerance):
        """Test whether this polyline is closed to within the tolerance.

        Args:
            tolerance: The minimum differnce between vertices below which vertices
                are considered the same.
        """
        return self._vertices[0].is_equivalent(self._vertices[-1], tolerance)

    def remove_colinear_vertices(self, tolerance):
        """Get a version of this polyline without colinear or duplicate vertices.

        Args:
            tolerance: The minimum distance that a vertex can be from a line
                before it is considered colinear.
        """
        if len(self.vertices) == 3:
            return self  # Polyline2D cannot have fewer than 3 vertices
        new_vertices = [self.vertices[0]]  # first vertex is always ok
        for i, _v in enumerate(self.vertices[1:-1]):
            _a = self[i].determinant(_v) + _v.determinant(self[i + 2]) + \
                self[i + 2].determinant(self[i])
            if abs(_a) >= tolerance:
                new_vertices.append(_v)
        new_vertices.append(self[-1])  # last vertex is always ok
        _new_poly = Polyline2D(new_vertices)
        self._transfer_properties(_new_poly)
        return _new_poly

    def reverse(self):
        """Get a copy of this polyline where the vertices are reversed."""
        _new_poly = Polyline2D(tuple(pt for pt in reversed(self.vertices)))
        self._transfer_properties(_new_poly)
        return _new_poly

    def move(self, moving_vec):
        """Get a polyline that has been moved along a vector.

        Args:
            moving_vec: A Vector2D with the direction and distance to move the polyline.
        """
        _new_poly = Polyline2D(tuple(pt.move(moving_vec) for pt in self.vertices))
        self._transfer_properties(_new_poly)
        return _new_poly

    def rotate(self, angle, origin):
        """Get a polyline that is rotated counterclockwise by a certain angle.

        Args:
            angle: An angle for rotation in radians.
            origin: A Point2D for the origin around which the point will be rotated.
        """
        _new_poly = Polyline2D(tuple(pt.rotate(angle, origin) for pt in self.vertices))
        self._transfer_properties(_new_poly)
        return _new_poly

    def reflect(self, normal, origin):
        """Get a polyline reflected across a plane with the input normal and origin.

        Args:
            normal: A Vector2D representing the normal vector for the plane across
                which the polyline will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point2D representing the origin from which to reflect.
        """
        _new_poly = Polyline2D(tuple(pt.reflect(normal, origin) for pt in self.vertices))
        self._transfer_properties(_new_poly)
        return _new_poly

    def scale(self, factor, origin=None):
        """Scale a polyline by a factor from an origin point.

        Args:
            factor: A number representing how much the polyline should be scaled.
            origin: A Point2D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0).
        """
        if origin is None:
            _new_poly = Polyline2D(tuple(
                Point2D(pt.x * factor, pt.y * factor) for pt in self.vertices))
        else:
            _new_poly = Polyline2D(tuple(
                pt.scale(factor, origin) for pt in self.vertices))
        _new_poly._interpolated = self._interpolated
        return _new_poly

    def intersect_line_ray(self, line_ray):
        """Get the intersections between this polyline and a Ray2D or LineSegment2D.

        Args:
            line_ray: A LineSegment2D or Ray2D or to intersect.

        Returns:
            A list with Point2D objects for the intersections.
            List will be empty if no intersection exists.
        """
        intersections = []
        for _s in self.segments:
            inters = intersect_line2d(_s, line_ray)
            if inters is not None:
                intersections.append(inters)
        return intersections

    def intersect_line_infinite(self, ray):
        """Get the intersections between this polyline and a Ray2D extended infintiely.

        Args:
            ray: A Ray2D or to intersect. This will be extended in both
                directions infinetly for the intersection.

        Returns:
            A list with Point2D objects for the intersections.
            List will be empty if no intersection exists.
        """
        intersections = []
        for _s in self.segments:
            inters = intersect_line2d_infinite(_s, ray)
            if inters is not None:
                intersections.append(inters)
        return intersections

    def to_dict(self):
        """Get Polyline2D as a dictionary."""
        base = {'type': 'Polyline2D',
                'vertices': [pt.to_array() for pt in self.vertices]}
        if self.interpolated:
            base['interpolated'] = self.interpolated
        return base

    def to_array(self):
        """Get a list of lists whenre each sub-list represents a Point2D vetex."""
        return tuple(pt.to_array() for pt in self.vertices)

    def to_polygon(self, tolerance):
        """Get a Polygon2D derived from this object.
        
        If the polyline is closed to within the tolerance, the segments of this
        polyline and the resulting polygon will match. Otherwise, an extra
        LineSegment2D will be added to connect the start and end of the polyline.

        Args:
            tolerance: The minimum differnce between vertices below which vertices
                are considered the same.
        """
        if self.is_closed(tolerance):
            return Polygon2D(self._vertices[:-1])
        return Polygon2D(self._vertices)

    @staticmethod
    def join_segments(segments, tolerance):
        """Get an array of Polyline2Ds from a list of LineSegment2Ds.

        Args:
            segments: An array of LineSegment2D objects.
            tolerance: The minimum difference in X, Y, and Z values at which Point2Ds
                are considred equivalent. Segments with points that match within the
                tolerance will be joined.

        Returns:
            An array of Polyline2D and LineSegment2D objects assembled from the
            joined segments.
        """
        # group the vertices that make up polylines
        grouped_verts = _group_vertices(segments, tolerance)

        # create the Polyline2D and LineSegment2D objects
        joined_lines = []
        for v_list in grouped_verts:
            if len(v_list) == 2:
                joined_lines.append(LineSegment2D.from_end_points(v_list[0], v_list[1]))
            else:
                joined_lines.append(Polyline2D(v_list))
        return joined_lines

    def _transfer_properties(self, new_polyline):
        """Transfer properties from this polyline to a new polyline."""
        new_polyline._interpolated = self._interpolated
        new_polyline._length = self._length
        new_polyline._is_self_intersecting = self._is_self_intersecting

    def __copy__(self):
        return Polyline2D(self._vertices, self._interpolated)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return tuple(hash(pt) for pt in self._vertices) + (self._interpolated,)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Polyline2D) and self.__key() == other.__key()

    def __repr__(self):
        return 'Polyline2D ({} vertices)'.format(len(self))



# coding=utf-8
import pytest

from ladybug_geometry.geometry2d.polyline import Polyline2D
from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry3d.polyline import Polyline3D
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.line import LineSegment3D
from ladybug_geometry.geometry3d.plane import Plane

import math


def test_polyline3d_init():
    """Test the initalization of Polyline3D objects and basic properties."""
    pts = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    pline = Polyline3D(pts)

    str(pline)  # test the string representation of the polyline

    assert isinstance(pline.vertices, tuple)
    assert len(pline.vertices) == 4
    assert len(pline) == 4
    for point in pline:
        assert isinstance(point, Point3D)

    assert isinstance(pline.segments, tuple)
    assert len(pline.segments) == 3
    for seg in pline.segments:
        assert isinstance(seg, LineSegment3D)
        assert seg.length == 2

    assert pline.length == 6
    assert pline.vertices[0] == pline[0]

    p_array = pline.to_array()
    assert isinstance(p_array, tuple)
    assert len(p_array) == 4
    for arr in p_array:
        assert isinstance(p_array, tuple)
        assert len(arr) == 3
    pline_2 = Polyline3D.from_array(p_array)
    assert pline == pline_2


def test_equality():
    """Test the equality of Polyline3D objects."""
    pts = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    pts_2 = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0.1, 2))
    pline = Polyline3D(pts)
    pline_dup = pline.duplicate()
    pline_alt = Polyline3D(pts_2)

    assert pline is pline
    assert pline is not pline_dup
    assert pline == pline_dup
    assert hash(pline) == hash(pline_dup)
    assert pline != pline_alt
    assert hash(pline) != hash(pline_alt)


def test_to_from_polygon():
    """Test the from_polygon method."""
    pts_1 = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pline2d = Polyline2D(pts_1, interpolated=True)

    pline = Polyline3D.from_polyline2d(pline2d)
    assert isinstance(pline, Polyline3D)
    assert len(pline) == 4
    assert pline.interpolated

    new_pline2d = pline.to_polyline2d()
    assert isinstance(new_pline2d, Polyline2D)
    assert len(new_pline2d) == 4
    assert new_pline2d.interpolated


def test_polyline3d_to_from_dict():
    """Test the to/from dict of Polyline3D objects."""
    pts = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    pline = Polyline3D(pts)
    pline_dict = pline.to_dict()
    new_pline = Polyline3D.from_dict(pline_dict)
    assert isinstance(new_pline, Polyline3D)
    assert new_pline.to_dict() == pline_dict


def test_min_max_center():
    """Test the Polyline3D min, max and center."""
    pts = (Point3D(0, 0, 3), Point3D(2, 0, 3), Point3D(2, 2, 3), Point3D(0, 2, 3))
    pline = Polyline3D(pts)

    assert pline.min == Point3D(0, 0, 3)
    assert pline.max == Point3D(2, 2, 3)
    assert pline.center == Point3D(1, 1, 3)


def test_remove_colinear_vertices():
    """Test the remove_colinear_vertices method of Polyline3D."""
    pts_1 = (Point3D(0, 0, 3), Point3D(2, 0, 3), Point3D(2, 2, 3), Point3D(0, 2, 3))
    pts_2 = (Point3D(0, 0, 3), Point3D(1, 0, 3), Point3D(2, 0, 3), Point3D(2, 2, 3),
             Point3D(0, 2, 3))
    pline_1 = Polyline3D(pts_1)
    pline_2 = Polyline3D(pts_2)

    assert len(pline_1.remove_colinear_vertices(0.0001).vertices) == 4
    assert len(pline_2.remove_colinear_vertices(0.0001).vertices) == 4


def test_pline3d_duplicate():
    """Test the duplicate method of Polyline3D."""
    pts = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    pline = Polyline3D(pts)
    new_pline = pline.duplicate()

    for i, pt in enumerate(new_pline):
        assert pt == pts[i]

    assert pline.length == new_pline.length
    assert pline.vertices == new_pline.vertices


def test_reverse():
    """Test the reverse method."""
    pts_1 = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    pline = Polyline3D(pts_1)
    new_pline = pline.reverse()

    assert pline.length == new_pline.length
    assert pline.vertices == tuple(reversed(new_pline.vertices))


def test_move():
    """Test the Polyline3D move method."""
    pts = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    pline = Polyline3D(pts)

    vec_1 = Vector3D(2, 2, 2)
    new_pline = pline.move(vec_1)
    assert new_pline[0] == Point3D(2, 2, 2)
    assert new_pline[1] == Point3D(4, 2, 2)
    assert new_pline[2] == Point3D(4, 4, 2)
    assert new_pline[3] == Point3D(2, 4, 2)
    assert pline.length == new_pline.length


def test_scale():
    """Test the Polyline3D scale method."""
    pts_1 = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    pline_1 = Polyline3D(pts_1)
    pts_2 = (Point3D(1, 1), Point3D(2, 1), Point3D(2, 2), Point3D(1, 2))
    pline_2 = Polyline3D(pts_2)
    origin_1 = Point3D(2, 0)
    origin_2 = Point3D(1, 1)

    new_pline_1 = pline_1.scale(2, origin_1)
    assert new_pline_1[0] == Point3D(-2, 0)
    assert new_pline_1[1] == Point3D(2, 0)
    assert new_pline_1[2] == Point3D(2, 4)
    assert new_pline_1[3] == Point3D(-2, 4)
    assert new_pline_1.length == pline_1.length * 2

    new_pline_2 = pline_2.scale(2, origin_2)
    assert new_pline_2[0] == Point3D(1, 1)
    assert new_pline_2[1] == Point3D(3, 1)
    assert new_pline_2[2] == Point3D(3, 3)
    assert new_pline_2[3] == Point3D(1, 3)
    assert new_pline_2.length == pline_2.length * 2


def test_scale_world_origin():
    """Test the Polyline3D scale method with None origin."""
    pts = (Point3D(1, 1, 2), Point3D(2, 1, 2), Point3D(2, 2, 2), Point3D(1, 2, 2))
    pline = Polyline3D(pts)

    new_pline = pline.scale(2)
    assert new_pline[0] == Point3D(2, 2, 4)
    assert new_pline[1] == Point3D(4, 2, 4)
    assert new_pline[2] == Point3D(4, 4, 4)
    assert new_pline[3] == Point3D(2, 4, 4)
    assert new_pline.length == pline.length * 2


def test_rotate():
    """Test the Polyline3D rotate method."""
    pts = (Point3D(1, 1, 2), Point3D(2, 1, 2), Point3D(2, 2, 2), Point3D(1, 2, 2))
    pline = Polyline3D(pts)
    origin_1 = Point3D(1, 1)

    test_1 = pline.rotate(Vector3D(0, 0, 1), math.pi, origin_1)
    assert test_1[0].x == pytest.approx(1, rel=1e-3)
    assert test_1[0].y == pytest.approx(1, rel=1e-3)
    assert test_1[2].x == pytest.approx(0, rel=1e-3)
    assert test_1[2].y == pytest.approx(0, rel=1e-3)
    assert pline.length == pytest.approx(test_1.length, rel=1e-3)

    test_2 = pline.rotate(Vector3D(0, 0, 1), math.pi/2, origin_1)
    assert test_2[0].x == pytest.approx(1, rel=1e-3)
    assert test_2[0].y == pytest.approx(1, rel=1e-3)
    assert test_2[2].x == pytest.approx(0, rel=1e-3)
    assert test_2[2].y == pytest.approx(2, rel=1e-3)


def test_rotate_xy():
    """Test the Polyline3D rotate_xy method."""
    pts = (Point3D(1, 1, 2), Point3D(2, 1, 2), Point3D(2, 2, 2), Point3D(1, 2, 2))
    pline = Polyline3D(pts)
    origin_1 = Point3D(1, 1)

    test_1 = pline.rotate_xy(math.pi, origin_1)
    assert test_1[0].x == pytest.approx(1, rel=1e-3)
    assert test_1[0].y == pytest.approx(1, rel=1e-3)
    assert test_1[2].x == pytest.approx(0, rel=1e-3)
    assert test_1[2].y == pytest.approx(0, rel=1e-3)
    assert pline.length == pytest.approx(test_1.length, rel=1e-3)

    test_2 = pline.rotate_xy(math.pi/2, origin_1)
    assert test_2[0].x == pytest.approx(1, rel=1e-3)
    assert test_2[0].y == pytest.approx(1, rel=1e-3)
    assert test_2[2].x == pytest.approx(0, rel=1e-3)
    assert test_2[2].y == pytest.approx(2, rel=1e-3)


def test_reflect():
    """Test the Polyline3D reflect method."""
    pts = (Point3D(1, 1, 2), Point3D(2, 1, 2), Point3D(2, 2, 2), Point3D(1, 2, 2))
    pline = Polyline3D(pts)

    origin_1 = Point3D(1, 0)
    normal_1 = Vector3D(1, 0)
    normal_2 = Vector3D(-1, -1).normalize()

    test_1 = pline.reflect(normal_1, origin_1)
    assert test_1[0].x == pytest.approx(1, rel=1e-3)
    assert test_1[0].y == pytest.approx(1, rel=1e-3)
    assert test_1[2].x == pytest.approx(0, rel=1e-3)
    assert test_1[2].y == pytest.approx(2, rel=1e-3)
    assert pline.length == pytest.approx(test_1.length, rel=1e-3)


def test_intersect_plane():
    """Test the Polyline3D intersect_plane method."""
    pts = (Point3D(0, 0, 1), Point3D(2, 0, -1), Point3D(2, 2, 1), Point3D(0, 2, -1))
    pline = Polyline3D(pts)

    plane_1 = Plane(n=Vector3D(0, 0, 1), o=Point3D(0, 0, 0))
    assert len(pline.intersect_plane(plane_1)) == 3

    plane_2 = Plane(n=Vector3D(0, 1, 0), o=Point3D(1, 1, 0))
    assert len(pline.intersect_plane(plane_2)) == 1

    plane_3 = Plane(n=Vector3D(0, 0, 1), o=Point3D(0, 0, 3))
    plane_4 = Plane(n=Vector3D(0, 0, 1), o=Point3D(0, 0, -2))
    assert len(pline.intersect_plane(plane_3)) == 0
    assert len(pline.intersect_plane(plane_4)) == 0


def test_split_with_plane():
    """Test the Polyline3D split_with_plane method."""
    pts = (Point3D(0, 0, 1), Point3D(2, 0, -1), Point3D(2, 2, 1), Point3D(0, 2, -1))
    pline = Polyline3D(pts)

    plane_1 = Plane(n=Vector3D(0, 0, 1), o=Point3D(0, 0, 0))
    splits = pline.split_with_plane(plane_1)
    assert len(splits) == 4
    assert isinstance(splits[0], LineSegment3D)
    assert isinstance(splits[1], Polyline3D)

    plane_2 = Plane(n=Vector3D(0, 1, 0), o=Point3D(1, 1, 0))
    assert len(pline.split_with_plane(plane_2)) == 2

    plane_3 = Plane(n=Vector3D(0, 0, 1), o=Point3D(0, 0, 3))
    plane_4 = Plane(n=Vector3D(0, 0, 1), o=Point3D(0, 0, -2))
    assert len(pline.split_with_plane(plane_3)) == 1
    assert len(pline.split_with_plane(plane_4)) == 1


def test_join_segments():
    """Test the join_segments method."""
    pts = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    l_segs = (LineSegment3D.from_end_points(pts[0], pts[1]),
            LineSegment3D.from_end_points(pts[1], pts[2]),
            LineSegment3D.from_end_points(pts[2], pts[3]),
            LineSegment3D.from_end_points(pts[3], pts[0]))
    p_lines = Polyline3D.join_segments(l_segs, 0.01)
    assert len(p_lines) == 1
    assert isinstance(p_lines[0], Polyline3D)
    assert len(p_lines[0]) == 5
    assert p_lines[0].is_closed(0.01)

    l_segs = (LineSegment3D.from_end_points(pts[0], pts[1]),
            LineSegment3D.from_end_points(pts[2], pts[3]),
            LineSegment3D.from_end_points(pts[1], pts[2]),
            LineSegment3D.from_end_points(pts[3], pts[0]))
    p_lines = Polyline3D.join_segments(l_segs, 0.01)
    assert len(p_lines) == 1
    assert isinstance(p_lines[0], Polyline3D)
    assert len(p_lines[0]) == 5
    assert p_lines[0].is_closed(0.01)

    l_segs = (LineSegment3D.from_end_points(pts[0], pts[1]),
            LineSegment3D.from_end_points(pts[1], pts[2]),
            LineSegment3D.from_end_points(pts[0], pts[3]),
            LineSegment3D.from_end_points(pts[3], pts[2]))
    p_lines = Polyline3D.join_segments(l_segs, 0.01)
    assert len(p_lines) == 1
    assert isinstance(p_lines[0], Polyline3D)
    assert len(p_lines[0]) == 5
    assert p_lines[0].is_closed(0.01)


def test_join_segments_multiple_pline():
    """Test the join_segments method with multiple polylines."""
    pts = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    extra_pts = (Point3D(3, 3), Point3D(4, 3), Point3D(4, 4), Point3D(3, 4))
    l_segs = (LineSegment3D.from_end_points(extra_pts[0], extra_pts[1]),
              LineSegment3D.from_end_points(pts[0], pts[1]),
              LineSegment3D.from_end_points(pts[1], pts[2]),
              LineSegment3D.from_end_points(pts[0], pts[3]),
              LineSegment3D.from_end_points(pts[3], pts[2]))
    p_lines = Polyline3D.join_segments(l_segs, 0.01)
    assert len(p_lines) == 2

    l_segs = (LineSegment3D.from_end_points(extra_pts[0], extra_pts[1]),
              LineSegment3D.from_end_points(pts[0], pts[1]),
              LineSegment3D.from_end_points(pts[1], pts[2]),
              LineSegment3D.from_end_points(pts[0], pts[3]),
              LineSegment3D.from_end_points(pts[3], pts[2]),
              LineSegment3D.from_end_points(extra_pts[2], extra_pts[3]),
              LineSegment3D.from_end_points(extra_pts[1], extra_pts[2]),
              LineSegment3D.from_end_points(extra_pts[0], extra_pts[3]))
    p_lines = Polyline3D.join_segments(l_segs, 0.01)
    assert len(p_lines) == 2
    for p_line in p_lines:
        assert isinstance(p_line, Polyline3D)
        assert len(p_line) == 5
        assert p_line.is_closed(0.01)


def test_join_segments_disconnected():
    """Test the join_segments method with diconnected polylines."""
    pts = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    extra_pts = (Point3D(3, 3), Point3D(4, 3), Point3D(4, 4), Point3D(3, 4))

    l_segs = (LineSegment3D.from_end_points(extra_pts[0], extra_pts[1]),
              LineSegment3D.from_end_points(pts[0], pts[1]),
              LineSegment3D.from_end_points(pts[2], pts[3]),
              LineSegment3D.from_end_points(extra_pts[3], extra_pts[2]))
    p_lines = Polyline3D.join_segments(l_segs, 0.01)
    assert len(p_lines) == 4



# coding=utf-8
import pytest

from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.cone import Cone

import math


def test_cone_init():
    """Test the initalization of Cone objects and basic properties."""
    vertex = Point3D(2, 0, 2)
    axis = Vector3D(0, 2, 2)
    angle = 0.7
    c = Cone(vertex, axis, angle)

    str(c)  # test the string representation of the cone
    assert c.vertex == Point3D(2, 0, 2)
    assert c.axis == Vector3D(0, 2, 2)
    assert c.angle == 0.7
    assert c.height == c.axis.magnitude
    assert isinstance(c.slant_height, float)
    assert isinstance(c.area, float)
    assert isinstance(c.volume, float)


def test_equality():
    """Test the equality of Cone objects."""
    cone = Cone(Point3D(2, 0, 2), Vector3D(0, 2, 2), 0.7)
    cone_dup = cone.duplicate()
    cone_alt = Cone(Point3D(2, 0.1, 2), Vector3D(0, 2, 2), 0.7)

    assert cone is cone
    assert cone is not cone_dup
    assert cone == cone_dup
    assert hash(cone) == hash(cone_dup)
    assert cone != cone_alt
    assert hash(cone) != hash(cone_alt)


def test_cone_to_from_dict():
    """Test the Cone to_dict and from_dict methods."""
    c = Cone(Point3D(4, 0.5, 2), Vector3D(1, 0, 2.5), 0.7)
    d = c.to_dict()
    c = Cone.from_dict(d)
    assert c.vertex.x == pytest.approx(4, rel=1e-3)
    assert c.vertex.y == pytest.approx(0.5, rel=1e-3)
    assert c.vertex.z == pytest.approx(2, rel=1e-3)
    assert c.axis.x == pytest.approx(1, rel=1e-3)
    assert c.axis.y == pytest.approx(0, rel=1e-3)
    assert c.axis.z == pytest.approx(2.5, rel=1e-3)


def test_cone_duplicate():
    """Test the Cone duplicate method."""
    c = Cone(Point3D(0, 0.5, 2), Vector3D(1, 0.5, 2.5), 0.75)
    test = c.duplicate()
    assert test.angle == 0.75
    assert c.vertex.x == pytest.approx(0, rel=1e-3)
    assert c.vertex.y == pytest.approx(0.5, rel=1e-3)
    assert c.vertex.z == pytest.approx(2, rel=1e-3)
    assert c.axis.x == pytest.approx(1, rel=1e-3)
    assert c.axis.y == pytest.approx(0.5, rel=1e-3)
    assert c.axis.z == pytest.approx(2.5, rel=1e-3)


def test_cone_rotate():
    """Test the Cone rotate method."""
    c = Cone(Point3D(2, 0, 2), Vector3D(1, 0.5, 2.5), 0.75)
    test1 = c.rotate(Vector3D(0, 0, 1), math.pi, Point3D(0, 0, 0))
    assert test1.vertex.x == pytest.approx(-2, rel=1e-3)
    assert test1.vertex.y == pytest.approx(0, rel=1e-3)
    assert test1.vertex.z == pytest.approx(2, rel=1e-3)
    assert test1.axis == Vector3D(1, 0.5, 2.5).rotate(Vector3D(0, 0, 1), math.pi)


def test_cone_rotate_xy():
    """Test the Cone rotate_xy method."""
    c = Cone(Point3D(1, 0, 3), Vector3D(1, 2, 0), 0.5)
    test = c.rotate_xy(math.pi / 2, Point3D(0, 0, 0))
    assert test.vertex.x == pytest.approx(0, rel=1e-3)
    assert test.vertex.y == pytest.approx(1, rel=1e-3)
    assert test.vertex.z == pytest.approx(3, rel=1e-3)
    assert test.axis == Vector3D(1, 2, 0).rotate_xy(math.pi / 2)


def test_cone_reflect():
    """Test the Cone reflect method."""
    origin_1 = Point3D(1, 0, 0)
    origin_2 = Point3D(0, 0, 2)
    normal_1 = Vector3D(0, 0, 1)
    normal_2 = Vector3D(1, 0, 0)

    c = Cone(Point3D(0, 0, 0), Vector3D(1, 2, 0), 0.5)
    test_1 = c.reflect(normal_1, origin_1)
    assert test_1.vertex.x == pytest.approx(0, rel=1e-3)
    assert test_1.vertex.y == pytest.approx(0, rel=1e-3)
    assert test_1.vertex.z == pytest.approx(0, rel=1e-3)

    test_2 = c.reflect(normal_2, origin_1)
    assert test_2.vertex.x == pytest.approx(2, rel=1e-3)
    assert test_2.vertex.y == pytest.approx(0, rel=1e-3)
    assert test_2.vertex.z == pytest.approx(0, rel=1e-3)

    test_3 = c.reflect(normal_1, origin_2)
    assert test_3.vertex.x == pytest.approx(0, rel=1e-3)
    assert test_3.vertex.y == pytest.approx(0, rel=1e-3)
    assert test_3.vertex.z == pytest.approx(4, rel=1e-3)


def test_cone_move():
    """Test the Cone move method."""
    c = Cone(Point3D(4, 0.5, 2), Vector3D(1, 0, 2.5), 0.7)
    test = c.move(Vector3D(2, 3, 6.5))
    assert test.vertex.x == pytest.approx(6, rel=1e-3)
    assert test.vertex.y == pytest.approx(3.5, rel=1e-3)
    assert test.vertex.z == pytest.approx(8.5, rel=1e-3)
    assert test.axis.x == pytest.approx(1, rel=1e-3)
    assert test.axis.y == pytest.approx(0, rel=1e-3)
    assert test.axis.z == pytest.approx(2.5, rel=1e-3)


def test_cone_scale():
    """Test the Cone scale method."""
    c = Cone(Point3D(4, 0.5, 2), Vector3D(1, 0, 2.5), 0.7)
    test = c.scale(2, Point3D(0, 0, 0))
    assert test.vertex.x == pytest.approx(8, rel=1e-3)
    assert test.vertex.y == pytest.approx(1, rel=1e-3)
    assert test.vertex.z == pytest.approx(4, rel=1e-3)
    assert test.axis.x == pytest.approx(2, rel=1e-3)
    assert test.axis.y == pytest.approx(0, rel=1e-3)
    assert test.axis.z == pytest.approx(5, rel=1e-3)



# coding=utf-8
"""Base class for 2D geometries in sD space (Surface3D and Mesh3D)."""
from __future__ import division

from .pointvector import Point3D


class Base2DIn3D(object):
    """Base class for 2D geometries in 3D space (Surface3D and Mesh3D).

    Args:
        vertices: A list of Point3D objects representing the vertices.

    Properties:
        * vertices
        * min
        * max
        * center
    """
    __slots__ = ('_vertices', '_min', '_max', '_center')

    def __init__(self, vertices):
        """Initilize Base2DIn3D."""
        self._vertices = self._check_vertices_input(vertices)
        self._min = None
        self._max = None
        self._center = None

    @property
    def vertices(self):
        """Tuple of all vertices in this object."""
        return self._vertices

    @property
    def min(self):
        """A Point3D for the minimum bounding box vertex around this geometry."""
        if self._min is None:
            self._calculate_min_max()
        return self._min

    @property
    def max(self):
        """A Point3D for the maximum bounding box vertex around this geometry."""
        if self._max is None:
            self._calculate_min_max()
        return self._max

    @property
    def center(self):
        """A Point3D for the center of the bounding box around this geometry."""
        if self._center is None:
            min, max = self.min, self.max
            self._center = Point3D(
                (min.x + max.x) / 2, (min.y + max.y) / 2, (min.z + max.z) / 2)
        return self._center

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def _calculate_min_max(self):
        """Calculate maximum and minimum Point3D for this object."""
        min_pt = [self.vertices[0].x, self.vertices[0].y, self.vertices[0].z]
        max_pt = [self.vertices[0].x, self.vertices[0].y, self.vertices[0].z]

        for v in self.vertices[1:]:
            if v.x < min_pt[0]:
                min_pt[0] = v.x
            elif v.x > max_pt[0]:
                max_pt[0] = v.x
            if v.y < min_pt[1]:
                min_pt[1] = v.y
            elif v.y > max_pt[1]:
                max_pt[1] = v.y
            if v.z < min_pt[2]:
                min_pt[2] = v.z
            elif v.z > max_pt[2]:
                max_pt[2] = v.z

        self._min = Point3D(min_pt[0], min_pt[1], min_pt[2])
        self._max = Point3D(max_pt[0], max_pt[1], max_pt[2])

    def _check_vertices_input(self, vertices):
        if not isinstance(vertices, tuple):
            vertices = tuple(vertices)
        assert len(vertices) >= 3, 'There must be at least 3 vertices for a {}.' \
            ' Got {}'.format(self.__class__.__name__, len(vertices))
        for vert in vertices:
            assert isinstance(vert, Point3D), \
                'Expected Point3D for {} vertex. Got {}.'.format(
                    self.__class__.__name__, type(vert))
        return vertices

    def __len__(self):
        return len(self._vertices)

    def __getitem__(self, key):
        return self._vertices[key]

    def __iter__(self):
        return iter(self._vertices)

    def __copy__(self):
        return Base2DIn3D(self._vertices)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return tuple(hash(pt) for pt in self._vertices)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Base2DIn3D) and self.__key() == other.__key()
    
    def __ne__(self, other):
        return not self.__eq__(other)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Base2Din3D representation."""
        return 'Base 2D Object (3D Space)'



# coding=utf-8
import pytest

from ladybug_geometry.geometry2d.pointvector import Vector2D
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry3d.line import LineSegment3D
from ladybug_geometry.geometry3d.ray import Ray3D
from ladybug_geometry.geometry3d.face import Face3D

import math


def test_face3d_init():
    """Test the initalization of Face3D objects and basic properties."""
    pts = (Point3D(0, 0, 2), Point3D(0, 2, 2), Point3D(2, 2, 2), Point3D(2, 0, 2))
    plane = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face = Face3D(pts, plane)
    str(face)  # test the string representation of the face

    assert isinstance(face.plane, Plane)
    assert face.plane.n == Vector3D(0, 0, 1)
    assert face.plane.n == face.normal
    assert face.plane.o == Point3D(0, 0, 2)

    assert isinstance(face.vertices, tuple)
    assert len(face.vertices) == 4
    assert len(face) == 4
    for point in face:
        assert isinstance(point, Point3D)

    assert isinstance(face.boundary_segments, tuple)
    assert len(face.boundary_segments) == 4
    for seg in face.boundary_segments:
        assert isinstance(seg, LineSegment3D)
        assert seg.length == 2
    assert face.has_holes is False
    assert face.hole_segments is None

    assert face.area == 4
    assert face.perimeter == 8
    assert face.is_clockwise is False
    assert face.is_convex is True
    assert face.is_self_intersecting is False
    assert face.vertices[0] == face[0]


def test_equality():
    """Test the equality of Face3D objects."""
    pts = (Point3D(0, 0, 2), Point3D(0, 2, 2), Point3D(2, 2, 2), Point3D(2, 0, 2))
    pts_2 = (Point3D(0.1, 0, 2), Point3D(0, 2, 2), Point3D(2, 2, 2), Point3D(2, 0, 2))
    plane = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face = Face3D(pts, plane)
    face_dup = face.duplicate()
    face_alt = Face3D(pts_2, plane)

    assert face is face
    assert face is not face_dup
    assert face == face_dup
    assert hash(face) == hash(face_dup)
    assert face != face_alt
    assert hash(face) != hash(face_alt)


def test_face3d_to_from_dict():
    """Test the to/from dict of Face3D objects."""
    pts = (Point3D(0, 0, 2), Point3D(0, 2, 2), Point3D(2, 2, 2), Point3D(2, 0, 2))
    plane = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face = Face3D(pts, plane)
    face_dict = face.to_dict()
    new_face = Face3D.from_dict(face_dict)
    assert isinstance(new_face, Face3D)
    assert new_face.to_dict() == face_dict

    bound_pts = [Point3D(0, 0), Point3D(4, 0), Point3D(4, 4), Point3D(0, 4)]
    hole_pts = [Point3D(1, 1), Point3D(3, 1), Point3D(3, 3), Point3D(1, 3)]
    face = Face3D(bound_pts, None, [hole_pts])
    face_dict = face.to_dict()
    new_face = Face3D.from_dict(face_dict)
    assert isinstance(new_face, Face3D)
    assert new_face.to_dict() == face_dict


def test_face3d_init_from_vertices():
    """Test the initalization of Face3D objects without a plane."""
    pts = (Point3D(0, 0, 2), Point3D(0, 2, 2), Point3D(2, 2, 2), Point3D(2, 0, 2))
    face = Face3D(pts)

    assert isinstance(face.plane, Plane)
    assert face.plane.n == Vector3D(0, 0, -1)
    assert face.plane.n == face.normal
    assert face.plane.o == Point3D(0, 0, 2)

    assert isinstance(face.vertices, tuple)
    assert len(face.vertices) == 4
    assert len(face) == 4
    for point in face:
        assert isinstance(point, Point3D)

    assert isinstance(face.boundary_segments, tuple)
    assert len(face.boundary_segments) == 4
    for seg in face.boundary_segments:
        assert isinstance(seg, LineSegment3D)
        assert seg.length == 2
    assert face.has_holes is False
    assert face.hole_segments is None

    assert face.area == 4
    assert face.perimeter == 8
    assert face.is_clockwise is False
    assert face.is_convex is True
    assert face.is_self_intersecting is False
    assert face.vertices[0] == face[0]


def test_face3d_init_from_vertices_colinear():
    """Test the initalization of Face3D objects with colinear vertices."""
    pts = (Point3D(0, 0, 2), Point3D(0, 1, 2), Point3D(0, 2, 2), Point3D(2, 2, 2),
           Point3D(2, 0, 2))
    face = Face3D(pts)

    assert not face.normal.is_zero()
    assert face.plane.n == Vector3D(0, 0, -1)
    assert face.plane.n == face.normal
    assert face.plane.o == Point3D(0, 0, 2)


def test_face3d_init_from_extrusion():
    """Test the initalization of Face3D from_extrusion."""
    line_seg = LineSegment3D(Point3D(0, 0, 0), Vector3D(2, 0, 0))
    extru_vec = Vector3D(0, 0, 2)
    face = Face3D.from_extrusion(line_seg, extru_vec)

    assert isinstance(face.plane, Plane)
    assert face.plane.is_coplanar(Plane(Vector3D(0, 1, 0), Point3D(0, 0, 0)))
    assert face.plane.n == face.normal
    assert face.plane.o == Point3D(0, 0, 0)

    assert isinstance(face.vertices, tuple)
    assert len(face.vertices) == 4
    assert len(face) == 4
    for point in face:
        assert isinstance(point, Point3D)

    assert isinstance(face.boundary_segments, tuple)
    assert len(face.boundary_segments) == 4
    for seg in face.boundary_segments:
        assert isinstance(seg, LineSegment3D)
        assert seg.length == 2
    assert face.has_holes is False
    assert face.hole_segments is None

    assert face.area == 4
    assert face.perimeter == 8
    assert face.is_clockwise is False
    assert face.is_convex is True
    assert face.is_self_intersecting is False


def test_face3d_init_from_rectangle():
    """Test the initalization of Face3D from_rectangle."""
    plane = Plane(Vector3D(0, 0, 1), Point3D(2, 2, 2))
    face = Face3D.from_rectangle(2, 2, plane)

    assert isinstance(face.plane, Plane)
    assert face.plane.n == Vector3D(0, 0, 1)
    assert face.plane.n == face.normal
    assert face.plane.o == Point3D(2, 2, 2)

    assert isinstance(face.vertices, tuple)
    assert len(face.vertices) == 4
    for point in face.vertices:
        assert isinstance(point, Point3D)

    assert isinstance(face.boundary_segments, tuple)
    assert len(face.boundary_segments) == 4
    for seg in face.boundary_segments:
        assert isinstance(seg, LineSegment3D)
        assert seg.length == 2
    assert face.has_holes is False
    assert face.hole_segments is None

    assert face.area == 4
    assert face.perimeter == 8
    assert face.is_clockwise is False
    assert face.is_convex is True
    assert face.is_self_intersecting is False


def test_face3d_init_from_regular_polygon():
    """Test the initalization of Face3D from_regular_polygon."""
    plane = Plane(Vector3D(0, 0, 1), Point3D(2, 2, 2))
    face = Face3D.from_regular_polygon(8, 2, plane)

    assert isinstance(face.plane, Plane)
    assert face.plane.n == Vector3D(0, 0, 1)
    assert face.plane.n == face.normal
    assert face.plane.o == Point3D(2, 2, 2)

    assert isinstance(face.vertices, tuple)
    assert len(face.vertices) == 8
    for point in face.vertices:
        assert isinstance(point, Point3D)
    assert isinstance(face.boundary_segments, tuple)
    assert len(face.boundary_segments) == 8
    for seg in face.boundary_segments:
        assert isinstance(seg, LineSegment3D)
        assert seg.length == pytest.approx(1.5307337, rel=1e-3)
    assert face.has_holes is False
    assert face.hole_segments is None

    assert face.area == pytest.approx(11.3137084, rel=1e-3)
    assert face.perimeter == pytest.approx(1.5307337 * 8, rel=1e-3)
    assert face.is_clockwise is False
    assert face.is_convex is True
    assert face.is_self_intersecting is False

    polygon = Face3D.from_regular_polygon(3)
    assert len(polygon.vertices) == 3
    polygon = Face3D.from_regular_polygon(20)
    assert len(polygon.vertices) == 20
    with pytest.raises(AssertionError):
        polygon = Face3D.from_regular_polygon(2)


def test_face3d_init_from_shape_with_hole():
    """Test the initalization of Face3D from_shape_with_holes with one hole."""
    bound_pts = [Point3D(0, 0), Point3D(4, 0), Point3D(4, 4), Point3D(0, 4)]
    hole_pts = [Point3D(1, 1), Point3D(3, 1), Point3D(3, 3), Point3D(1, 3)]
    face = Face3D(bound_pts, None, [hole_pts])

    assert face.plane.n == Vector3D(0, 0, 1)
    assert face.plane.n == face.normal
    assert face.plane.o == Point3D(0, 0, 0)

    assert isinstance(face.vertices, tuple)
    assert len(face.vertices) == 10
    for point in face.vertices:
        assert isinstance(point, Point3D)

    assert isinstance(face.boundary_segments, tuple)
    assert len(face.boundary_segments) == 4
    for seg in face.boundary_segments:
        assert isinstance(seg, LineSegment3D)
    assert face.has_holes is True
    assert isinstance(face.hole_segments, tuple)
    assert len(face.hole_segments) == 1
    assert len(face.hole_segments[0]) == 4
    for seg in face.hole_segments[0]:
        assert isinstance(seg, LineSegment3D)

    assert face.area == 12
    assert face.perimeter == pytest.approx(24, rel=1e-3)
    assert face.is_clockwise is False
    assert face.is_convex is False
    assert face.is_self_intersecting is False


def test_face3d_init_from_shape_with_holes():
    """Test the initalization of Face3D from_shape_with_holes."""
    bound_pts = [Point3D(0, 0), Point3D(4, 0), Point3D(4, 4), Point3D(0, 4)]
    hole_pts_1 = [Point3D(1, 1), Point3D(1.5, 1), Point3D(1.5, 1.5), Point3D(1, 1.5)]
    hole_pts_2 = [Point3D(2, 2), Point3D(3, 2), Point3D(3, 3), Point3D(2, 3)]
    face = Face3D(bound_pts, None, [hole_pts_1, hole_pts_2])

    assert face.plane.n == Vector3D(0, 0, 1)
    assert face.plane.n == face.normal
    assert face.plane.o == Point3D(0, 0, 0)

    assert isinstance(face.vertices, tuple)
    assert len(face.vertices) == 16
    for point in face.vertices:
        assert isinstance(point, Point3D)

    assert isinstance(face.boundary_segments, tuple)
    assert len(face.boundary_segments) == 4
    for seg in face.boundary_segments:
        assert isinstance(seg, LineSegment3D)
    assert face.has_holes is True
    assert isinstance(face.hole_segments, tuple)
    assert len(face.hole_segments) == 2

    assert face.area == 16 - 1.25
    assert face.perimeter == pytest.approx(22, rel=1e-3)
    assert face.is_clockwise is False
    assert face.is_convex is False
    assert face.is_self_intersecting is False


def test_face3d_init_from_punched_geometry():
    """Test the initalization of Face3D from_shape_with_holes."""
    bound_pts = [Point3D(0, 0), Point3D(4, 0), Point3D(4, 4), Point3D(0, 4)]
    hole_pts_1 = [Point3D(1, 1), Point3D(1.5, 1), Point3D(1.5, 1.5), Point3D(1, 1.5)]
    hole_pts_2 = [Point3D(2, 2), Point3D(3, 2), Point3D(3, 3), Point3D(2, 3)]
    face_1 = Face3D(bound_pts)
    face_2 = Face3D(bound_pts, None, [hole_pts_1])
    sub_face = Face3D(hole_pts_2)

    face = Face3D.from_punched_geometry(face_1, [sub_face])
    assert isinstance(face.vertices, tuple)
    assert len(face.vertices) == 10
    assert face.has_holes is True
    assert isinstance(face.hole_segments, tuple)
    assert len(face.hole_segments) == 1
    assert face.area == 15
    assert face.perimeter == pytest.approx(20, rel=1e-3)
    assert face.is_clockwise is False
    assert face.is_convex is False
    assert face.is_self_intersecting is False

    face = Face3D.from_punched_geometry(face_2, [sub_face])
    assert isinstance(face.vertices, tuple)
    assert len(face.vertices) == 16
    assert face.has_holes is True
    assert isinstance(face.hole_segments, tuple)
    assert len(face.hole_segments) == 2
    assert face.area == 16 - 1.25
    assert face.perimeter == pytest.approx(22, rel=1e-3)
    assert face.is_clockwise is False
    assert face.is_convex is False
    assert face.is_self_intersecting is False


def test_is_geometrically_equivalent():
    """Test the is_geometrically_equivalent method."""
    plane_1 = Plane(Vector3D(0, 0, 1))
    plane_2 = Plane(Vector3D(0, 0, -1))
    pts_1 = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    pts_2 = (Point3D(0, 0), Point3D(0, 2), Point3D(2, 2), Point3D(2, 0))
    pts_3 = (Point3D(0, 0), Point3D(2, 0), Point3D(2.1, 2.1), Point3D(0, 2))
    pts_4 = (Point3D(1, 0), Point3D(0, 1), Point3D(1, 2), Point3D(2, 1))
    pts_5 = (Point3D(0, 0), Point3D(2, 2), Point3D(2, 0), Point3D(0, 2))
    pts_6 = (Point3D(2, 0), Point3D(2, 2), Point3D(0, 2), Point3D(0, 0))
    face_1 = Face3D(pts_1, plane_1)
    face_2 = Face3D(pts_2, plane_1)
    face_3 = Face3D(pts_1, plane_2)
    face_4 = Face3D(pts_2, plane_2)
    face_5 = Face3D(pts_3, plane_1)
    face_6 = Face3D(pts_4, plane_1)
    face_7 = Face3D(pts_5, plane_1)
    face_8 = Face3D(pts_6, plane_1)

    assert face_1.is_geometrically_equivalent(face_2, 0.0001) is True
    assert face_1.is_geometrically_equivalent(face_3, 0.0001) is True
    assert face_1.is_geometrically_equivalent(face_4, 0.0001) is True
    assert face_1.is_geometrically_equivalent(face_5, 0.0001) is False
    assert face_1.is_geometrically_equivalent(face_6, 0.0001) is False
    assert face_1.is_geometrically_equivalent(face_7, 0.0001) is False
    assert face_1.is_geometrically_equivalent(face_8, 0.0001) is True


def test_is_centered_adjacent():
    """Test the is_centered_adjacent method."""
    plane_1 = Plane(Vector3D(0, 0, 1))
    plane_2 = Plane(Vector3D(0, 0, -1))
    pts_1 = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    pts_2 = (Point3D(0, 0), Point3D(0, 2), Point3D(2, 2), Point3D(2, 0))
    pts_3 = (Point3D(0, 0), Point3D(2, 0), Point3D(2.1, 2.1), Point3D(0, 2))
    pts_4 = (Point3D(1, 0), Point3D(0, 1), Point3D(1, 2), Point3D(2, 1))
    pts_5 = (Point3D(2, 0), Point3D(2, 2), Point3D(0, 2), Point3D(0, 0))
    face_1 = Face3D(pts_1, plane_1)
    face_2 = Face3D(pts_2, plane_1)
    face_3 = Face3D(pts_1, plane_2)
    face_4 = Face3D(pts_2, plane_2)
    face_5 = Face3D(pts_3, plane_1)
    face_6 = Face3D(pts_4, plane_1)
    face_7 = Face3D(pts_5, plane_1)

    assert face_1.is_centered_adjacent(face_2, 0.0001) is True
    assert face_1.is_centered_adjacent(face_3, 0.0001) is True
    assert face_1.is_centered_adjacent(face_4, 0.0001) is True
    assert face_1.is_centered_adjacent(face_5, 0.0001) is False
    assert face_1.is_centered_adjacent(face_6, 0.0001) is False
    assert face_1.is_centered_adjacent(face_7, 0.0001) is True


def test_is_sub_face():
    """Test the is_sub_face method."""
    bound_pts = [Point3D(0, 0), Point3D(4, 0), Point3D(4, 4), Point3D(0, 4)]
    sub_pts_1 = [Point3D(1, 1), Point3D(1.5, 1), Point3D(1.5, 1.5), Point3D(1, 1.5)]
    sub_pts_2 = [Point3D(2, 2), Point3D(3, 2), Point3D(3, 3), Point3D(2, 3)]
    sub_pts_3 = [Point3D(2, 2), Point3D(6, 2), Point3D(6, 6), Point3D(2, 6)]
    sub_pts_4 = [Point3D(5, 5), Point3D(6, 5), Point3D(6, 6), Point3D(5, 6)]
    sub_pts_5 = [Point3D(), Point3D(2), Point3D(2, 0, 2), Point3D(0, 0, 2)]
    plane_1 = Plane(Vector3D(0, 0, 1))
    plane_2 = Plane(Vector3D(0, 0, -1))
    plane_3 = Plane(Vector3D(0, 1, 0))
    face = Face3D(bound_pts, plane_1)
    sub_face_1 = Face3D(sub_pts_1, plane_1)
    sub_face_2 = Face3D(sub_pts_2, plane_2)
    sub_face_3 = Face3D(sub_pts_3, plane_1)
    sub_face_4 = Face3D(sub_pts_4, plane_1)
    sub_face_5 = Face3D(sub_pts_5, plane_3)

    assert face.is_sub_face(sub_face_1, 0.0001, 0.0001) is True
    assert face.is_sub_face(sub_face_2, 0.0001, 0.0001) is True
    assert face.is_sub_face(sub_face_3, 0.0001, 0.0001) is False
    assert face.is_sub_face(sub_face_4, 0.0001, 0.0001) is False
    assert face.is_sub_face(sub_face_5, 0.0001, 0.0001) is False


def test_clockwise():
    """Test the clockwise property."""
    plane_1 = Plane(Vector3D(0, 0, 1))
    plane_2 = Plane(Vector3D(0, 0, -1))
    pts_1 = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    pts_2 = (Point3D(0, 0), Point3D(0, 2), Point3D(2, 2), Point3D(2, 0))
    face_1 = Face3D(pts_1, plane_1, enforce_right_hand=True)
    face_2 = Face3D(pts_2, plane_1, enforce_right_hand=True)
    face_3 = Face3D(pts_1, plane_2, enforce_right_hand=True)
    face_4 = Face3D(pts_2, plane_2, enforce_right_hand=True)

    assert face_1.is_clockwise is face_1.polygon2d.is_clockwise is False
    assert face_2.is_clockwise is face_2.polygon2d.is_clockwise is False
    assert face_3.is_clockwise is face_3.polygon2d.is_clockwise is False
    assert face_4.is_clockwise is face_4.polygon2d.is_clockwise is False

    face_1 = Face3D(pts_1, plane_1, enforce_right_hand=False)
    face_2 = Face3D(pts_2, plane_1, enforce_right_hand=False)
    face_3 = Face3D(pts_1, plane_2, enforce_right_hand=False)
    face_4 = Face3D(pts_2, plane_2, enforce_right_hand=False)

    assert face_1.is_clockwise is face_1.polygon2d.is_clockwise is False
    assert face_2.is_clockwise is face_2.polygon2d.is_clockwise is True
    assert face_3.is_clockwise is face_3.polygon2d.is_clockwise is True
    assert face_4.is_clockwise is face_4.polygon2d.is_clockwise is False

    assert face_1.area == face_2.area == face_3.area == face_4.area == 4


def test_is_convex():
    """Test the convex property."""
    plane_1 = Plane(Vector3D(0, 0, 1))
    plane_2 = Plane(Vector3D(0, 0, -1))
    pts_1 = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    pts_2 = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 1), Point3D(1, 1),
             Point3D(1, 2), Point3D(0, 2))
    face_1 = Face3D(pts_1, plane_1)
    face_2 = Face3D(pts_2, plane_1)
    face_3 = Face3D(pts_1, plane_2)
    face_4 = Face3D(pts_2, plane_2)

    assert face_1.is_convex is True
    assert face_2.is_convex is False
    assert face_3.is_convex is True
    assert face_4.is_convex is False


def test_is_self_intersecting():
    """Test the is_self_intersecting property."""
    plane_1 = Plane(Vector3D(0, 0, 1))
    plane_2 = Plane(Vector3D(0, 0, -1))
    pts_1 = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    pts_2 = (Point3D(0, 0), Point3D(0, 2), Point3D(2, 0), Point3D(2, 2))
    face_1 = Face3D(pts_1, plane_1)
    face_2 = Face3D(pts_2, plane_1)
    face_3 = Face3D(pts_1, plane_2)
    face_4 = Face3D(pts_2, plane_2)

    assert face_1.is_self_intersecting is False
    assert face_2.is_self_intersecting is True
    assert face_3.is_self_intersecting is False
    assert face_4.is_self_intersecting is True


def test_is_valid():
    """Test the is_valid property."""
    plane_1 = Plane(Vector3D(0, 0, 1))
    pts_1 = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2))
    pts_2 = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 0))
    face_1 = Face3D(pts_1, plane_1)
    face_2 = Face3D(pts_2, plane_1)

    assert face_1.is_valid is True
    assert face_2.is_valid is False


def test_min_max_center():
    """Test the Face3D min, max and center."""
    pts_1 = (Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 2, 2), Point3D(0, 2, 2))
    pts_2 = (Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(0, 0, 2))
    plane_1 = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    plane_2 = Plane(Vector3D(0, 1, 0))
    face_1 = Face3D(pts_1, plane_1)
    face_2 = Face3D(pts_2, plane_2)

    assert face_1.min == Point3D(0, 0, 2)
    assert face_1.max == Point3D(2, 2, 2)
    assert face_1.center == Point3D(1, 1, 2)

    assert face_2.min == Point3D(0, 0, 0)
    assert face_2.max == Point3D(2, 0, 2)
    assert face_2.center == Point3D(1, 0, 1)


def test_upper_left_counter_clockwise_vertices():
    """Test the upper_left_counter_clockwise_vertices property."""
    plane_1 = Plane(Vector3D(0, 1, 0))
    plane_2 = Plane(Vector3D(0, -1, 0))
    pts_1 = (Point3D(0, 0, 0), Point3D(2, 0, 0), Point3D(2, 0, 2), Point3D(0, 0, 2))
    pts_2 = (Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 0, 0))
    face_1 = Face3D(pts_1, plane_1, enforce_right_hand=False)
    face_2 = Face3D(pts_2, plane_1, enforce_right_hand=False)
    face_3 = Face3D(pts_1, plane_2, enforce_right_hand=False)
    face_4 = Face3D(pts_2, plane_2, enforce_right_hand=False)

    up_cclock_1 = face_1.upper_left_counter_clockwise_vertices
    assert up_cclock_1[0] == Point3D(2, 0, 2)
    assert face_1.is_clockwise is True
    assert Face3D(up_cclock_1, enforce_right_hand=False).is_clockwise is False
    up_cclock_2 = face_2.upper_left_counter_clockwise_vertices
    assert up_cclock_2[0] == Point3D(2, 0, 2)
    assert face_2.is_clockwise is False
    assert not Face3D(up_cclock_2, enforce_right_hand=False).is_clockwise
    assert up_cclock_1 == up_cclock_2

    up_cclock_3 = face_3.upper_left_counter_clockwise_vertices
    assert up_cclock_3[0] == Point3D(0, 0, 2)
    assert face_3.is_clockwise is False
    assert Face3D(up_cclock_3, enforce_right_hand=False).is_clockwise is False
    up_cclock_4 = face_4.upper_left_counter_clockwise_vertices
    assert up_cclock_4[0] == Point3D(0, 0, 2)
    assert face_4.is_clockwise is True
    assert Face3D(up_cclock_4, enforce_right_hand=False).is_clockwise is False


def test_upper_left_counter_clockwise_vertices_triangle():
    """Test the upper_left_counter_clockwise_vertices property with triangles."""
    plane_1 = Plane(Vector3D(0, 1, 0))
    pts_1 = (Point3D(0, 0, 0), Point3D(2, 0, -1), Point3D(0, 0, 2))
    pts_2 = (Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, -1))
    face_1 = Face3D(pts_1, plane_1, enforce_right_hand=False)
    face_2 = Face3D(pts_2, plane_1, enforce_right_hand=False)

    up_cclock_1 = face_1.upper_left_counter_clockwise_vertices
    assert up_cclock_1[0] == Point3D(0, 0, 2)
    assert face_1.is_clockwise is True
    assert Face3D(up_cclock_1, enforce_right_hand=False).is_clockwise is False
    up_cclock_2 = face_2.upper_left_counter_clockwise_vertices
    assert up_cclock_2[0] == Point3D(0, 0, 2)
    assert face_2.is_clockwise is False
    assert not Face3D(up_cclock_2, enforce_right_hand=False).is_clockwise
    assert up_cclock_1 == up_cclock_2


def test_duplicate():
    """Test the duplicate method of Face3D."""
    pts = (Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 2, 2), Point3D(0, 2, 2))
    plane_1 = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face = Face3D(pts, plane_1)
    new_face = face.duplicate()

    for i, pt in enumerate(new_face):
        assert pt == pts[i]

    assert face.area == new_face.area
    assert face.perimeter == new_face.perimeter
    assert face.is_clockwise == new_face.is_clockwise
    assert face.is_convex == new_face.is_convex
    assert face.is_self_intersecting == new_face.is_self_intersecting


def test_remove_colinear_vertices():
    """Test the remove_colinear_vertices method of Face3D."""
    pts_1 = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 2), Point3D(0, 2))
    pts_2 = (Point3D(0, 0), Point3D(1, 0), Point3D(2, 0), Point3D(2, 2),
             Point3D(0, 2))
    plane_1 = Plane(Vector3D(0, 0, 1))
    face_1 = Face3D(pts_1, plane_1)
    face_2 = Face3D(pts_2, plane_1)

    assert len(face_1.remove_colinear_vertices(0.0001).vertices) == 4
    assert len(face_2.remove_colinear_vertices(0.0001).vertices) == 4


def test_triangulated_mesh_and_centroid():
    """Test the duplicate method of Face3D."""
    pts_1 = (Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 2, 2), Point3D(0, 2, 2))
    plane_1 = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face_1 = Face3D(pts_1, plane_1)

    assert len(face_1.triangulated_mesh3d.vertices) == 4
    assert len(face_1.triangulated_mesh3d.faces) == 2
    assert face_1.triangulated_mesh3d.area == 4

    assert face_1.triangulated_mesh3d.min == Point3D(0, 0, 2)
    assert face_1.triangulated_mesh3d.max == Point3D(2, 2, 2)
    assert face_1.triangulated_mesh3d.center == Point3D(1, 1, 2)
    assert face_1.centroid == Point3D(1, 1, 2)

    pts_2 = (Point3D(0, 0), Point3D(0, 2), Point3D(2, 2),
             Point3D(4, 0), Point3D(2, -2))
    plane_2 = Plane(Vector3D(0, 0, 1))
    face_2 = Face3D(pts_2, plane_2)

    assert len(face_2.triangulated_mesh3d.vertices) == 5
    assert len(face_2.triangulated_mesh3d.faces) == 3
    assert face_2.triangulated_mesh3d.area == 10

    assert face_2.triangulated_mesh3d.min == Point3D(0, -2, 0)
    assert face_2.triangulated_mesh3d.max == Point3D(4, 2, 0)
    assert face_2.triangulated_mesh3d.center == Point3D(2, 0, 0)
    assert face_2.centroid.x == pytest.approx(1.73, rel=1e-2)
    assert face_2.centroid.y == pytest.approx(0.2667, rel=1e-2)
    assert face_2.centroid.z == 0


def test_check_planar():
    """Test the check_planar method of Face3D."""
    pts_1 = (Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 2, 2), Point3D(0, 2, 2))
    pts_2 = (Point3D(0, 0, 0), Point3D(2, 0, 2), Point3D(2, 2, 2), Point3D(0, 2, 2))
    pts_3 = (Point3D(0, 0, 2.0001), Point3D(2, 0, 2), Point3D(2, 2, 2), Point3D(0, 2, 2))
    plane_1 = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face_1 = Face3D(pts_1, plane_1)
    face_2 = Face3D(pts_2, plane_1)
    face_3 = Face3D(pts_3, plane_1)

    assert face_1.check_planar(0.001) is True
    assert face_2.check_planar(0.001, False) is False
    with pytest.raises(Exception):
        face_2.check_planar(0.0001)
    assert face_3.check_planar(0.001) is True
    assert face_3.check_planar(0.000001, False) is False
    with pytest.raises(Exception):
        face_3.check_planar(0.000001)


def test_flip():
    """Test the flip method of Face3D."""
    pts_1 = (Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 2, 2), Point3D(0, 2, 2))
    plane_1 = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face_1 = Face3D(pts_1, plane_1)
    face_2 = face_1.flip()

    assert face_1.normal == face_2.normal.reverse()
    assert face_1.is_clockwise is False
    assert face_2.is_clockwise is False
    for i, pt in enumerate(reversed(face_1.vertices)):
        assert pt == face_2[i]


def test_move():
    """Test the Face3D move method."""
    pts_1 = (Point3D(0, 0, 0), Point3D(2, 0, 0), Point3D(2, 2, 0), Point3D(0, 2, 0))
    plane_1 = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 0))
    face_1 = Face3D(pts_1, plane_1)

    vec_1 = Vector3D(2, 2, 2)
    new_face = face_1.move(vec_1)
    assert new_face[0] == Point3D(2, 2, 2)
    assert new_face[1] == Point3D(4, 2, 2)
    assert new_face[2] == Point3D(4, 4, 2)
    assert new_face[3] == Point3D(2, 4, 2)
    assert new_face.plane.o == face_1.plane.o.move(vec_1)
    assert new_face.plane.n == face_1.plane.n

    assert face_1.area == new_face.area
    assert face_1.perimeter == new_face.perimeter
    assert face_1.is_clockwise is new_face.is_clockwise
    assert face_1.is_convex is new_face.is_convex
    assert face_1.is_self_intersecting is new_face.is_self_intersecting
    assert new_face.normal == face_1.normal


def test_scale():
    """Test the Face3D scale method."""
    pts_1 = (Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 2, 2), Point3D(0, 2, 2))
    plane_1 = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face_1 = Face3D(pts_1, plane_1)
    pts_2 = (Point3D(1, 1), Point3D(2, 1), Point3D(2, 2), Point3D(1, 2))
    plane_2 = Plane(Vector3D(0, 0, 1))
    face_2 = Face3D(pts_2, plane_2)
    origin_1 = Point3D(2, 0)
    origin_2 = Point3D(1, 1)

    new_face_1 = face_1.scale(2, origin_1)
    assert new_face_1[0] == Point3D(-2, 0, 4)
    assert new_face_1[1] == Point3D(2, 0, 4)
    assert new_face_1[2] == Point3D(2, 4, 4)
    assert new_face_1[3] == Point3D(-2, 4, 4)
    assert new_face_1.area == face_1.area * 2 ** 2
    assert new_face_1.perimeter == face_1.perimeter * 2
    assert new_face_1.is_clockwise is face_1.is_clockwise
    assert new_face_1.is_convex is face_1.is_convex
    assert new_face_1.is_self_intersecting is face_1.is_self_intersecting
    assert new_face_1.normal == face_1.normal

    new_face_2 = face_2.scale(2, origin_2)
    assert new_face_2[0] == Point3D(1, 1)
    assert new_face_2[1] == Point3D(3, 1)
    assert new_face_2[2] == Point3D(3, 3)
    assert new_face_2[3] == Point3D(1, 3)
    assert new_face_2.area == face_2.area * 2 ** 2
    assert new_face_2.perimeter == face_2.perimeter * 2
    assert new_face_2.is_clockwise is face_2.is_clockwise
    assert new_face_2.is_convex is face_2.is_convex
    assert new_face_2.is_self_intersecting is face_2.is_self_intersecting
    assert new_face_2.normal == face_2.normal


def test_scale_world_origin():
    """Test the Face3D scale method with None origin."""
    pts = (Point3D(1, 1, 2), Point3D(2, 1, 2), Point3D(2, 2, 2), Point3D(1, 2, 2))
    plane = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face = Face3D(pts, plane)

    new_face = face.scale(2)
    assert new_face[0] == Point3D(2, 2, 4)
    assert new_face[1] == Point3D(4, 2, 4)
    assert new_face[2] == Point3D(4, 4, 4)
    assert new_face[3] == Point3D(2, 4, 4)
    assert new_face.area == face.area * 2 ** 2
    assert new_face.perimeter == face.perimeter * 2
    assert new_face.is_clockwise is face.is_clockwise is False
    assert new_face.is_convex is face.is_convex
    assert new_face.is_self_intersecting is face.is_self_intersecting
    assert new_face.normal == face.normal

    face = Face3D(pts, plane, enforce_right_hand=False)
    new_face = face.scale(2)
    assert new_face.is_clockwise is False
    new_face = face.scale(-2)
    assert new_face.is_clockwise is False


def test_rotate():
    """Test the Face3D rotate method."""
    pts = (Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 2, 2), Point3D(0, 2, 2))
    plane = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face = Face3D(pts, plane)
    origin = Point3D(0, 0, 0)
    axis = Vector3D(1, 0, 0)

    test_1 = face.rotate(axis, math.pi, origin)
    assert test_1[0].x == pytest.approx(0, rel=1e-3)
    assert test_1[0].y == pytest.approx(0, rel=1e-3)
    assert test_1[0].z == pytest.approx(-2, rel=1e-3)
    assert test_1[2].x == pytest.approx(2, rel=1e-3)
    assert test_1[2].y == pytest.approx(-2, rel=1e-3)
    assert test_1[2].z == pytest.approx(-2, rel=1e-3)
    assert face.area == test_1.area
    assert len(face.vertices) == len(test_1.vertices)

    test_2 = face.rotate(axis, math.pi/2, origin)
    assert test_2[0].x == pytest.approx(0, rel=1e-3)
    assert test_2[0].y == pytest.approx(-2, rel=1e-3)
    assert test_2[0].z == pytest.approx(0, rel=1e-3)
    assert test_2[2].x == pytest.approx(2, rel=1e-3)
    assert test_2[2].y == pytest.approx(-2, rel=1e-3)
    assert test_2[2].z == pytest.approx(2, rel=1e-3)
    assert face.area == test_2.area
    assert len(face.vertices) == len(test_2.vertices)

    face = Face3D(pts, plane, enforce_right_hand=False)
    test_1 = face.rotate(axis, math.pi, origin)
    assert test_1.is_clockwise is False
    test_2 = face.rotate(axis, math.pi/2, origin)
    assert test_2.is_clockwise is False


def test_rotate_xy():
    """Test the Face3D rotate_xy method."""
    pts = (Point3D(1, 1, 2), Point3D(2, 1, 2), Point3D(2, 2, 2), Point3D(1, 2, 2))
    plane = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face = Face3D(pts, plane)
    origin_1 = Point3D(1, 1, 0)

    test_1 = face.rotate_xy(math.pi, origin_1)
    assert test_1[0].x == pytest.approx(1, rel=1e-3)
    assert test_1[0].y == pytest.approx(1, rel=1e-3)
    assert test_1[0].z == pytest.approx(2, rel=1e-3)
    assert test_1[2].x == pytest.approx(0, rel=1e-3)
    assert test_1[2].y == pytest.approx(0, rel=1e-3)
    assert test_1[2].z == pytest.approx(2, rel=1e-3)

    test_2 = face.rotate_xy(math.pi/2, origin_1)
    assert test_2[0].x == pytest.approx(1, rel=1e-3)
    assert test_2[0].y == pytest.approx(1, rel=1e-3)
    assert test_1[0].z == pytest.approx(2, rel=1e-3)
    assert test_2[2].x == pytest.approx(0, rel=1e-3)
    assert test_2[2].y == pytest.approx(2, rel=1e-3)
    assert test_1[2].z == pytest.approx(2, rel=1e-3)

    face = Face3D(pts, plane, enforce_right_hand=False)
    test_1 = face.rotate_xy(math.pi, origin_1)
    assert test_1.is_clockwise is False
    test_2 = face.rotate_xy(math.pi/2, origin_1)
    assert test_2.is_clockwise is False


def test_reflect():
    """Test the Face3D reflect method."""
    pts = (Point3D(1, 1, 2), Point3D(2, 1, 2), Point3D(2, 2, 2), Point3D(1, 2, 2))
    plane = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face = Face3D(pts, plane)

    origin_1 = Point3D(1, 0, 2)
    normal_1 = Vector3D(1, 0, 0)
    normal_2 = Vector3D(-1, -1, 0).normalize()

    test_1 = face.reflect(normal_1, origin_1)
    assert test_1[-1].x == pytest.approx(1, rel=1e-3)
    assert test_1[-1].y == pytest.approx(1, rel=1e-3)
    assert test_1[-1].z == pytest.approx(2, rel=1e-3)
    assert test_1[1].x == pytest.approx(0, rel=1e-3)
    assert test_1[1].y == pytest.approx(2, rel=1e-3)
    assert test_1[1].z == pytest.approx(2, rel=1e-3)

    test_1 = face.reflect(normal_2, Point3D(0, 0, 2))
    assert test_1.is_clockwise is False
    assert test_1[-1].x == pytest.approx(-1, rel=1e-3)
    assert test_1[-1].y == pytest.approx(-1, rel=1e-3)
    assert test_1[-1].z == pytest.approx(2, rel=1e-3)
    assert test_1[1].x == pytest.approx(-2, rel=1e-3)
    assert test_1[1].y == pytest.approx(-2, rel=1e-3)
    assert test_1[1].z == pytest.approx(2, rel=1e-3)

    test_2 = face.reflect(normal_2, origin_1)
    assert test_2.is_clockwise is False
    assert test_2[-1].x == pytest.approx(0, rel=1e-3)
    assert test_2[-1].y == pytest.approx(0, rel=1e-3)
    assert test_2[-1].z == pytest.approx(2, rel=1e-3)
    assert test_2[1].x == pytest.approx(-1, rel=1e-3)
    assert test_2[1].y == pytest.approx(-1, rel=1e-3)
    assert test_2[1].z == pytest.approx(2, rel=1e-3)

    face = Face3D(pts, plane, enforce_right_hand=False)
    test_1 = face.reflect(normal_1, origin_1)
    assert test_1.is_clockwise is False
    test_1 = face.reflect(normal_2, Point3D(0, 0, 2))
    assert test_1.is_clockwise is False
    test_2 = face.reflect(normal_2, origin_1)
    assert test_2.is_clockwise is False


def test_intersect_line_ray():
    """Test the Face3D intersect_line_ray method."""
    pts = (Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 1, 2), Point3D(1, 1, 2),
           Point3D(1, 2, 2), Point3D(0, 2, 2))
    plane = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face = Face3D(pts, plane)
    ray_1 = Ray3D(Point3D(0.5, 0.5, 0), Vector3D(0, 0, 1))
    ray_2 = Ray3D(Point3D(0.5, 0.5, 0), Vector3D(0, 0, -1))
    ray_3 = Ray3D(Point3D(1.5, 1.5, 0), Vector3D(0, 0, 1))
    ray_4 = Ray3D(Point3D(-1, -1, 0), Vector3D(0, 0, 1))
    line_1 = LineSegment3D(Point3D(0.5, 0.5, 0), Vector3D(0, 0, 3))
    line_2 = LineSegment3D(Point3D(0.5, 0.5, 0), Vector3D(0, 0, 1))

    assert face.intersect_line_ray(ray_1) == Point3D(0.5, 0.5, 2)
    assert face.intersect_line_ray(ray_2) is None
    assert face.intersect_line_ray(ray_3) is None
    assert face.intersect_line_ray(ray_4) is None
    assert face.intersect_line_ray(line_1) == Point3D(0.5, 0.5, 2)
    assert face.intersect_line_ray(line_2) is None


def test_intersect_plane():
    """Test the Face3D intersect_plane method."""
    pts = (Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 1, 2), Point3D(1, 1, 2),
           Point3D(1, 2, 2), Point3D(0, 2, 2))
    plane = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face = Face3D(pts, plane)

    plane_1 = Plane(Vector3D(0, 1, 0), Point3D(0.5, 0.5, 0))
    plane_2 = Plane(Vector3D(1, 0, 0), Point3D(0.5, 0.5, 0))
    plane_3 = Plane(Vector3D(0, 1, 0), Point3D(0.5, 1.5, 0))
    plane_4 = Plane(Vector3D(0, 1, 0), Point3D(0, 3, 0))
    plane_5 = Plane(Vector3D(1, 1, 0), Point3D(0, 2.5, 0))

    assert len(face.intersect_plane(plane_1)) == 1
    assert face.intersect_plane(plane_1)[0].p1 == Point3D(2, 0.5, 2)
    assert face.intersect_plane(plane_1)[0].p2 == Point3D(0, 0.5, 2)
    assert len(face.intersect_plane(plane_2)) == 1
    assert face.intersect_plane(plane_2)[0].p1 == Point3D(0.5, 0, 2)
    assert face.intersect_plane(plane_2)[0].p2 == Point3D(0.5, 2, 2)
    assert len(face.intersect_plane(plane_3)) == 1
    assert face.intersect_plane(plane_3)[0].p1 == Point3D(1, 1.5, 2)
    assert face.intersect_plane(plane_3)[0].p2 == Point3D(0, 1.5, 2)
    assert face.intersect_plane(plane_4) is None
    assert len(face.intersect_plane(plane_5)) == 2


def test_project_point():
    """Test the Face3D project_point method."""
    pts = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 1), Point3D(1, 1),
           Point3D(1, 2), Point3D(0, 2))
    plane = Plane(Vector3D(0, 0, 1))
    face = Face3D(pts, plane)

    pt_1 = Point3D(0.5, 0.5, 2)
    pt_2 = Point3D(0.5, 0.5, -2)
    pt_3 = Point3D(1.5, 1.5, 2)
    pt_4 = Point3D(-1, -1, 2)

    assert face.project_point(pt_1) == Point3D(0.5, 0.5, 0)
    assert face.project_point(pt_2) == Point3D(0.5, 0.5, 0)
    assert face.project_point(pt_3) is None
    assert face.project_point(pt_4) is None


def test_get_mesh_grid():
    """Test the Face3D get_mesh_grid method."""
    pts = (Point3D(0, 0), Point3D(0, 2), Point3D(2, 2), Point3D(2, 0))
    plane = Plane(Vector3D(0, 0, 1))
    face = Face3D(pts, plane)
    mesh = face.get_mesh_grid(0.5)

    assert len(mesh.vertices) == 25
    assert len(mesh.faces) == 16
    assert mesh.area == 4

    assert mesh.min.x == 0
    assert mesh.min.y == 0
    assert mesh.min.z == 0
    assert mesh.max.x == 2
    assert mesh.max.y == 2
    assert mesh.max.z == 0
    assert mesh.center.x == 1
    assert mesh.center.y == 1
    assert len(mesh.face_areas) == 16
    assert mesh.face_areas[0] == 0.25
    assert len(mesh.face_centroids) == 16

    mesh_2 = face.get_mesh_grid(0.5, 0.5, 1, False)
    mesh_3 = face.get_mesh_grid(0.5, 0.5, 1, True)

    assert mesh_2.min.z == pytest.approx(1, rel=1e-2)
    assert mesh_2.max.z == pytest.approx(1, rel=1e-2)
    assert mesh_3.min.z == pytest.approx(-1, rel=1e-2)
    assert mesh_3.max.z == pytest.approx(-1, rel=1e-2)


def test_countour_by_number():
    """Test the countour_by_number method."""
    pts_1 = [Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 0, 0)]
    pts_2 = [Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(4, 0, 0)]
    plane = Plane(Vector3D(0, 1, 0))
    face_1 = Face3D(pts_1, plane)
    face_2 = Face3D(pts_2, plane)

    contours = face_1.countour_by_number(4)
    assert len(contours) == 4
    assert contours[0].p2.z == pytest.approx(2, rel=1e-3)
    assert contours[-1].p2.z == pytest.approx(0.5, rel=1e-3)

    contours = face_1.countour_by_number(4, Vector2D(1))
    assert len(contours) == 4
    assert contours[-1].p2.x == pytest.approx(1.5, rel=1e-3)

    contours = face_1.countour_by_number(4, Vector2D(1), True)
    assert len(contours) == 4
    assert contours[-1].p2.x == pytest.approx(0.5, rel=1e-3)

    contours = face_2.countour_by_number(4)
    assert len(contours) == 4
    contours = face_2.countour_by_number(8, Vector2D(1))
    assert len(contours) == 8


def test_countour_by_distance_between():
    """Test the countour_by_distance_between method."""
    pts_1 = [Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 0, 0)]
    pts_2 = [Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(4, 0, 0)]
    plane = Plane(Vector3D(0, 1, 0))
    face_1 = Face3D(pts_1, plane)
    face_2 = Face3D(pts_2, plane)

    contours = face_1.countour_by_distance_between(0.5)
    assert len(contours) == 4
    assert contours[0].p2.z == pytest.approx(2, rel=1e-3)
    assert contours[-1].p2.z == pytest.approx(0.5, rel=1e-3)

    contours = face_1.countour_by_distance_between(0.5, Vector2D(1))
    assert len(contours) == 4
    assert contours[-1].p2.x == pytest.approx(1.5, rel=1e-3)

    contours = face_1.countour_by_distance_between(0.5, Vector2D(1), True)
    assert len(contours) == 4
    assert contours[-1].p2.x == pytest.approx(0.5, rel=1e-3)

    contours = face_2.countour_by_distance_between(0.5)
    assert len(contours) == 4
    contours = face_2.countour_by_distance_between(0.5, Vector2D(1))
    assert len(contours) == 8


def test_countour_fins_by_number():
    """Test the countour_fins_by_number method."""
    pts_1 = [Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 0, 0)]
    plane = Plane(Vector3D(0, 1, 0))
    face_1 = Face3D(pts_1, plane)

    fins = face_1.countour_fins_by_number(4, 0.5, 0.5)
    assert len(fins) == 4

    fins = face_1.countour_fins_by_number(4, 0.5, 0.5, contour_vector=Vector2D(1))
    assert len(fins) == 4

    fins = face_1.countour_fins_by_number(4, 0.5, 0.5, math.pi/4)
    assert len(fins) == 4


def test_countour_fins_by_distance_between():
    """Test the countour_fins_by_distance_between method."""
    pts_1 = [Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 0, 0)]
    pts_2 = [Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(4, 0, 0)]
    plane = Plane(Vector3D(0, 1, 0))
    face_1 = Face3D(pts_1, plane)
    face_2 = Face3D(pts_2, plane)

    fins = face_1.countour_fins_by_distance_between(0.5, 0.5, 0.5)
    assert len(fins) == 4

    fins = face_1.countour_fins_by_distance_between(0.25, 0.5, 0.5,
                                                    contour_vector=Vector2D(1))
    assert len(fins) == 8

    fins = face_2.countour_fins_by_distance_between(0.5, 0.5, 0.5,
                                                    contour_vector=Vector2D(1))
    assert len(fins) == 8


def test_extract_rectangle():
    """Test the Face3D extract_rectangle method."""
    pts_1 = [Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 0, 0)]
    pts_2 = [pt for pt in reversed(pts_1)]
    pts_3 = [Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(4, 0, 0)]
    pts_4 = [Point3D(-2, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(4, 0, 0)]
    pts_5 = [Point3D(0, 0, 0), Point3D(-2, 0, 2), Point3D(4, 0, 2), Point3D(2, 0, 0)]
    pts_6 = [Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(1, 0, 3), Point3D(2, 0, 2),
             Point3D(2, 0, 0)]
    pts_7 = [Point3D(-2, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2),
             Point3D(-1, 0, 0)]
    plane = Plane(Vector3D(0, 1, 0))
    face_1 = Face3D(pts_1, plane)
    face_2 = Face3D(pts_2, plane)
    face_3 = Face3D(pts_3, plane)
    face_4 = Face3D(pts_4, plane)
    face_5 = Face3D(pts_5, plane)
    face_6 = Face3D(pts_6, plane)
    face_7 = Face3D(pts_7, plane)

    f1_result = face_1.extract_rectangle(0.0001)
    assert f1_result[0] == LineSegment3D.from_end_points(Point3D(2, 0, 0), Point3D(0, 0, 0))
    assert f1_result[1] == LineSegment3D.from_end_points(Point3D(2, 0, 2), Point3D(0, 0, 2))
    assert len(f1_result[2]) == 0
    f2_result = face_2.extract_rectangle(0.0001)
    assert f2_result[0] == LineSegment3D.from_end_points(Point3D(2, 0, 0), Point3D(0, 0, 0))
    assert f2_result[1] == LineSegment3D.from_end_points(Point3D(2, 0, 2), Point3D(0, 0, 2))
    assert len(f2_result[2]) == 0
    f3_result = face_3.extract_rectangle(0.0001)
    assert f3_result[0] == LineSegment3D.from_end_points(Point3D(2, 0, 0), Point3D(0, 0, 0))
    assert f3_result[1] == LineSegment3D.from_end_points(Point3D(2, 0, 2), Point3D(0, 0, 2))
    assert len(f3_result[2]) == 1
    f4_result = face_4.extract_rectangle(0.0001)
    assert f4_result[0] == LineSegment3D.from_end_points(Point3D(2, 0, 0), Point3D(0, 0, 0))
    assert f4_result[1] == LineSegment3D.from_end_points(Point3D(2, 0, 2), Point3D(0, 0, 2))
    assert len(f4_result[2]) == 2
    f5_result = face_5.extract_rectangle(0.0001)
    assert f5_result[0] == LineSegment3D.from_end_points(Point3D(2, 0, 0), Point3D(0, 0, 0))
    assert f5_result[1] == LineSegment3D.from_end_points(Point3D(2, 0, 2), Point3D(0, 0, 2))
    assert len(f5_result[2]) == 2
    f6_result = face_6.extract_rectangle(0.0001)
    assert f6_result[0] == LineSegment3D.from_end_points(Point3D(2, 0, 0), Point3D(0, 0, 0))
    assert f6_result[1] == LineSegment3D.from_end_points(Point3D(2, 0, 2), Point3D(0, 0, 2))
    assert len(f6_result[2]) == 1
    f7_result = face_7.extract_rectangle(0.0001)
    assert f7_result is None


def test_extract_rectangle_complex():
    """Test the Face3D extract_rectangle method with a more complex shape."""
    pts_1 = (Point3D(-1, -1, 0), Point3D(-12, -1, 0), Point3D(-12, -1, 2),
             Point3D(-10, -1, 3), Point3D(-1, -1, 3))
    plane = Plane(Vector3D(0, 1, 0))
    face_1 = Face3D(pts_1, plane)
    f1_result = face_1.extract_rectangle(0.0001)

    assert f1_result[0] == LineSegment3D.from_end_points(Point3D(-1, -1, 0), Point3D(-10, -1, 0))
    assert f1_result[1] == LineSegment3D.from_end_points(Point3D(-1, -1, 3), Point3D(-10, -1, 3))
    assert len(f1_result[2]) == 1
    assert len(f1_result[2][0].vertices) == 4


def test_sub_faces_by_ratio_rectangle():
    """Test the Face3D sub_faces_by_ratio_rectangle method."""
    pts_1 = (Point3D(0, 0, 0), Point3D(12, 0, 0), Point3D(12, 0, 2), Point3D(0, 0, 3))
    plane_1 = Plane(Vector3D(0, 1, 0))
    plane_2 = Plane(Vector3D(0, -1, 0))
    face_1 = Face3D(pts_1, plane_1)
    face_2 = Face3D(pts_1, plane_2)

    sub_faces = face_1.sub_faces_by_ratio_rectangle(0.4, 0.0001)
    assert len(sub_faces) == 2
    areas = [srf.area for srf in sub_faces]
    assert sum(areas) == pytest.approx(face_1.area * 0.4, rel=1e-3)
    for face in sub_faces:
        assert face.is_clockwise is False

    sub_faces = face_2.sub_faces_by_ratio_rectangle(0.4, 0.0001)
    for face in sub_faces:
        assert face.is_clockwise is False


def test_sub_faces_by_ratio_sub_rectangle():
    """Test the Face3D sub_faces_by_ratio_sub_rectangle method."""
    pts_1 = (Point3D(0, 0, 0), Point3D(12, 0, 0), Point3D(12, 0, 2), Point3D(0, 0, 3))
    plane = Plane(Vector3D(0, 1, 0))
    face_1 = Face3D(pts_1, plane)
    window_height = 2
    sill_height = 1
    horiz_separation = 3
    vert_separation = 0

    sub_faces = face_1.sub_faces_by_ratio_sub_rectangle(
        0.4, window_height, sill_height, horiz_separation, vert_separation, 0.0001)
    assert len(sub_faces) == 5
    areas = [srf.area for srf in sub_faces]
    assert sum(areas) == pytest.approx(face_1.area * 0.4, rel=1e-3)

    sub_faces_2 = face_1.sub_faces_by_ratio_sub_rectangle(
        0.99, window_height, sill_height, horiz_separation, vert_separation, 0.0001)
    assert len(sub_faces_2) == 2
    areas = [srf.area for srf in sub_faces_2]
    assert sum(areas) == pytest.approx(face_1.area * 0.99, rel=1e-3)


def test_sub_faces_by_ratio():
    """Test the sub_faces_by_ratio method."""
    pts_1 = (Point3D(0, 0), Point3D(0, 2), Point3D(2, 2), Point3D(2, 0))
    pts_2 = (Point3D(0, 0), Point3D(2, 0), Point3D(2, 1), Point3D(1, 1),
             Point3D(1, 2), Point3D(0, 2))
    plane = Plane(Vector3D(0, 0, 1))
    face_1 = Face3D(pts_1, plane)
    face_2 = Face3D(pts_2, plane)

    sub_faces_1 = face_1.sub_faces_by_ratio(0.5)
    assert len(sub_faces_1) == 1
    assert sub_faces_1[0].area == pytest.approx(face_1.area * 0.5, rel=1e-3)

    sub_faces_2 = face_2.sub_faces_by_ratio(0.5)
    assert len(sub_faces_2) == 4
    areas = [srf.area for srf in sub_faces_2]
    assert sum(areas) == pytest.approx(face_2.area * 0.5, rel=1e-3)


def test_sub_faces_by_dimension_rectangle():
    """Test the sub_faces_by_dimension_rectangle method."""
    pts_1 = (Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(2, 0, 2), Point3D(2, 0, 0))
    pts_2 = (Point3D(0, 0, 0), Point3D(0, 0, 2), Point3D(4, 0, 2), Point3D(4, 0, 0))
    plane = Plane(Vector3D(0, 0, 1))
    face_1 = Face3D(pts_1)
    face_2 = Face3D(pts_2)
    sub_face_height = 1.0
    sill_height = 1.0
    rect_height = 2.0
    div_dist = 1.0

    sub_faces_1 = face_1.sub_faces_by_dimension_rectangle(
        sub_face_height, 3.0, sill_height, div_dist, 0.1)
    assert len(sub_faces_1) == 1
    segs_1 = sub_faces_1[0].boundary_segments
    assert segs_1[1].length == sub_face_height

    sub_faces_2 = face_2.sub_faces_by_dimension_rectangle(
        sub_face_height, 1.0, sill_height, div_dist, 0.1)
    assert len(sub_faces_2) == 3
    segs_2 = sub_faces_2[0].boundary_segments
    assert segs_2[0].length == 1.0
    assert segs_2[1].length == sub_face_height



# coding=utf-8
"""Utility functions for computing intersections between geometry in 2D space."""
from __future__ import division

from .geometry2d.pointvector import Point2D, Vector2D

import math


def intersect_line2d(line_ray_a, line_ray_b):
    """Get the intersection between any Ray2D or LineSegment2D objects as a Point2D.

    This function calculates scaling parameters for ua and ub where:
        A.p + ua * A.v = B.p + ub * B.v
    Which represents the intersection point between line A and line B.

    The derivation of ua is achieved by crossing both sides of the above equation
    with the direction vector of B, and rearranging the formula:
        A.p + ua * A.v = B.p + ub * B.v
        (A.p + ua * A.v) x B.v = (B.p + ub * B.v) x B.v # Cross both sides with B.v
        (A.p x B.v) + (ua * A.v x B.v) = (B.p x B.v) + (ub * B.v x B.v) # B.v x B.v = 0
        ua = (B.p - A.p) x B.v / (A.v x B.v)

    Args:
        line_ray_a: A LineSegment2D or Ray2D object.
        line_ray_b: Another LineSegment2D or Ray2D to intersect.

    Returns:
        Point2D of intersection if it exists. None if no intersection exists.
    """
    # d is the determinant between lines, if 0 lines are collinear
    d = line_ray_b.v.y * line_ray_a.v.x - line_ray_b.v.x * line_ray_a.v.y
    if d == 0:
        return None

    # (dx, dy) = A.p - B.p
    dy = line_ray_a.p.y - line_ray_b.p.y
    dx = line_ray_a.p.x - line_ray_b.p.x

    # Find parameters ua and ub for intersection between two lines

    # Calculate scaling parameter for line_ray_b
    ua = (line_ray_b.v.x * dy - line_ray_b.v.y * dx) / d
    # Checks the bounds of ua to ensure it obeys ray/line behaviour
    if not line_ray_a._u_in(ua):
        return None

    # Calculate scaling parameter for line_ray_b
    ub = (line_ray_a.v.x * dy - line_ray_a.v.y * dx) / d
    # Checks the bounds of ub to ensure it obeys ray/line behaviour
    if not line_ray_b._u_in(ub):
        return None

    return Point2D(line_ray_a.p.x + ua * line_ray_a.v.x,
                   line_ray_a.p.y + ua * line_ray_a.v.y)


def intersect_line2d_infinite(line_ray_a, line_ray_b):
    """Get the intersection between a Ray2D/LineSegment2D and another extended infinitely.

    Args:
        line_ray_a: A LineSegment2D or Ray2D object.
        line_ray_b: ALineSegment2D or Ray2D that will be extended infinitely
            for intersection.

    Returns:
        Point2D of intersection if it exists. None if no intersection exists.
    """
    d = line_ray_b.v.y * line_ray_a.v.x - line_ray_b.v.x * line_ray_a.v.y
    if d == 0:
        return None
    dy = line_ray_a.p.y - line_ray_b.p.y
    dx = line_ray_a.p.x - line_ray_b.p.x
    ua = (line_ray_b.v.x * dy - line_ray_b.v.y * dx) / d
    if not line_ray_a._u_in(ua):
        return None
    return Point2D(line_ray_a.p.x + ua * line_ray_a.v.x,
                   line_ray_a.p.y + ua * line_ray_a.v.y)


def does_intersection_exist_line2d(line_ray_a, line_ray_b):
    """Boolean denoting whether an intersection exists between Ray2D or LineSegment2D.

    This is slightly faster than actually computing the intersection but should only be
    used in cases where the actual point of intersection is not needed.

    Args:
        line_ray_a: A LineSegment2D or Ray2D object.
        line_ray_b: Another LineSegment2D or Ray2D to intersect.

    Returns:
        True if an intersection exists. False if no intersection exists.
    """
    d = line_ray_b.v.y * line_ray_a.v.x - line_ray_b.v.x * line_ray_a.v.y
    if d == 0:
        return False
    dy = line_ray_a.p.y - line_ray_b.p.y
    dx = line_ray_a.p.x - line_ray_b.p.x
    ua = (line_ray_b.v.x * dy - line_ray_b.v.y * dx) / d
    if not line_ray_a._u_in(ua):
        return False
    ub = (line_ray_a.v.x * dy - line_ray_a.v.y * dx) / d
    if not line_ray_b._u_in(ub):
        return False
    return True


def intersect_line2d_arc2d(line_ray, arc):
    """Get the intersection between any Ray2D/LineSegment2D and an Arc2D.

    Args:
        line_ray: A LineSegment2D or Ray2D object.
        arc: An Arc2D object along which the closest point will be determined.

    Returns:
        A list of 2 Point2D objects if a full intersection exists.
        A list with a single Point2D object if the line is tangent or intersects
        only once. None if no intersection exists.
    """
    a = line_ray.v.magnitude_squared
    b = 2 * (line_ray.v.x * (line_ray.p.x - arc.c.x) +
             line_ray.v.y * (line_ray.p.y - arc.c.y))
    c = arc.c.magnitude_squared + line_ray.p.magnitude_squared - \
        2 * arc.c.dot(line_ray.p) - arc.r ** 2
    det = b ** 2 - 4 * a * c
    if det < 0:
        return None
    sq = math.sqrt(det)
    u1 = (-b + sq) / (2 * a)
    u2 = (-b - sq) / (2 * a)
    pt1 = Point2D(line_ray.p.x + u1 * line_ray.v.x,
                  line_ray.p.y + u1 * line_ray.v.y) if line_ray._u_in(u1) else None
    pt2 = Point2D(line_ray.p.x + u2 * line_ray.v.x,
                  line_ray.p.y + u2 * line_ray.v.y) if line_ray._u_in(u2) else None

    if u1 == u2:  # Tangent
        pt = Point2D(line_ray.p.x + u1 * line_ray.v.x,
                     line_ray.p.y + u1 * line_ray.v.y)
        return pt if arc._pt_in(pt) else None

    pts = [p for p in (pt1, pt2) if p is not None and arc._pt_in(p)]
    return pts if len(pts) != 0 else None


def intersect_line2d_infinite_arc2d(line_ray, arc):
    """Get the intersection between an Arc2D and a Ray2D/LineSegment2D extended infinitely.

    Args:
        line_ray: A LineSegment2D or Ray2D that will be extended infinitely
            for intersection.
        arc: An Arc2D object along which the closest point will be determined.

    Returns:
        A list of 2 Point2D objects if a full intersection exists.
        A list with a single Point2D object if the line is tangent or intersects
        only once. None if no intersection exists.
    """
    a = line_ray.v.magnitude_squared
    b = 2 * (line_ray.v.x * (line_ray.p.x - arc.c.x) +
             line_ray.v.y * (line_ray.p.y - arc.c.y))
    c = arc.c.magnitude_squared + line_ray.p.magnitude_squared - \
        2 * arc.c.dot(line_ray.p) - arc.r ** 2
    det = b ** 2 - 4 * a * c
    if det < 0:
        return None
    sq = math.sqrt(det)
    u1 = (-b + sq) / (2 * a)
    u2 = (-b - sq) / (2 * a)

    if u1 == u2:  # Tangent
        pt = Point2D(line_ray.p.x + u1 * line_ray.v.x, line_ray.p.y + u1 * line_ray.v.y)
        return pt if arc._pt_in(pt) else None

    pt1 = Point2D(line_ray.p.x + u1 * line_ray.v.x, line_ray.p.y + u1 * line_ray.v.y)
    pt2 = Point2D(line_ray.p.x + u2 * line_ray.v.x, line_ray.p.y + u2 * line_ray.v.y)
    pts = [p for p in (pt1, pt2) if arc._pt_in(p)]
    return pts if len(pts) != 0 else None


def closest_point2d_on_line2d(point, line_ray):
    """Get the closest Point2D on a LineSegment2D or Ray2D to the input point.

    Args:
        point: A Point2D object.
        line_ray: A LineSegment2D or Ray2D object along which the closest point
            will be determined.

    Returns:
        Point2D for the closest point on the line_ray to point.
    """
    d = line_ray.v.magnitude_squared
    assert d != 0, '{} length must not equal 0.'.format(line_ray.__class__.__name__)
    u = ((point.x - line_ray.p.x) * line_ray.v.x +
         (point.y - line_ray.p.y) * line_ray.v.y) / d
    if not line_ray._u_in(u):
        u = max(min(u, 1.0), 0.0)
    return Point2D(line_ray.p.x + u * line_ray.v.x, line_ray.p.y + u * line_ray.v.y)

def closest_point2d_between_line2d(line_ray_a, line_ray_b):
    """Get the two closest Point2D between two LineSegment2D objects.

    Note that the line segments should not intersect for the result to be valid.

    Args:
        line_ray_a: A LineSegment2D object.
        line_ray_b: Another LineSegment2D to which closest points will
            be determined.

    Returns:
        A tuple with two elements

        - dists[0]: The distance between the two LineSegment2D objects.
        - pts[0]: A tuple of two Point2D objects representing:

        1) The point on line_ray_a that is closest to line_ray_b
        2) The point on line_ray_b that is closest to line_ray_a
    """
    # one of the 4 endpoints must be a closest point
    pt_1 = closest_point2d_on_line2d(line_ray_a.p, line_ray_b)
    dist_1 = pt_1.distance_to_point(line_ray_a.p)
    a_p2 = line_ray_a.p2
    pt_2 = closest_point2d_on_line2d(a_p2, line_ray_b)
    dist_2 = pt_2.distance_to_point(a_p2)
    pt_3 = closest_point2d_on_line2d(line_ray_b.p, line_ray_a)
    dist_3 = pt_3.distance_to_point(line_ray_b.p)
    b_p2 = line_ray_b.p2
    pt_4 = closest_point2d_on_line2d(b_p2, line_ray_a)
    dist_4 = pt_4.distance_to_point(b_p2)

    # sort the closest points based on their distance
    dists = [dist_1, dist_2, dist_3, dist_4]
    pts = [(line_ray_a.p, pt_1), (a_p2, pt_2), (pt_3, line_ray_b.p), (pt_4, b_p2)]
    dists, i = zip(*sorted(zip(dists, range(len(pts)))))
    return dists[0], pts[i[0]]


def closest_point2d_on_arc2d(point, arc):
    """Get the closest Point2D on a Arc2D to the input point.

    Args:
        point: A Point2D object.
        arc: An Arc2D object along which the closest point will be determined.

    Returns:
        Point2D for the closest point on arc to point.
    """
    v = point - arc.c
    v = v.normalize() * arc.r
    if arc.is_circle:
        return Point2D(arc.c.x + v.x, arc.c.y + v.y)
    else:
        a = Vector2D(1, 0).angle_counterclockwise(v)
        if (not arc.is_inverted and arc.a1 < a < arc.a2) or \
                (arc.is_inverted and arc.a1 > a > arc.a2):
            return Point2D(arc.c.x + v.x, arc.c.y + v.y)
        else:
            if arc.p1.distance_to_point(point) <= arc.p2.distance_to_point(point):
                return arc.p1
    return arc.p2
